﻿/*
 *
 *
 *    EndpointSMI.cs
 *
 *    A serial endpoint for SMI modems.  This is the main SMI protocol
 *    layer implementation, responsible for multiplexing commands and
 *    data over a single communications link, either serial or stream
 *    socket.
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2017 by Matthew K. Roberts, KK5JY. All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *    
 *
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using System.Threading;
using KK5JY.IO;

namespace KK5JY.SMI {
	/// <summary>
	/// An SMI communications endpoint.
	/// </summary>
	public class EndpointSMI : IStart, IDisposable {
		/// <summary>
		/// The largest command response to process.
		/// </summary>
		public const int MaximumCommandResponseSize = 4096;

		/// <summary>
		/// The value of Backspace key.
		/// </summary>
		private const int Backspace = 8;

		/// <summary>
		/// Initialization states.
		/// </summary>
		private enum InitStates {
			/// <summary>
			/// No data received yet from the modem.
			/// </summary>
			None,

			/// <summary>
			/// PING received, and VERSION sent.
			/// </summary>
			VersionSent,

			/// <summary>
			/// VERSION received, and ECHO sent.
			/// </summary>
			EchoSent,

			/// <summary>
			/// ECHO received, and MODE sent.
			/// </summary>
			ModeSent,

			/// <summary>
			/// MODE received, DOC sent.
			/// </summary>
			DocSent,

			/// <summary>
			/// ECHO received and modem ready.
			/// </summary>
			Ready,
		}

		#region Private Fields
		// the socket/serial-port
		private IDeviceHelper m_Device;
		// set to true if the device has reported a valid version string
		private InitStates m_State;
		// set to true while we are reading a command response
		private bool m_Escaped;
		// the received command buffer
		private string m_ResponseBuffer;
		// flag that controls consumption of extra CRLF after a command
		private byte m_LastByteWasProcessedCommand;
		// how often to retry the initialization sequence
		private readonly int m_InitRetryInterval = 5000;
		#endregion

		/// <summary>
		/// Constructor.
		/// </summary>
		/// <param name="device">The data link object.</param>
		public EndpointSMI(IDeviceHelper device) {
			m_Device = device;
			device.DataReceived += DeviceDataReceived;
			device.LinkStateChanged += LinkStateChangeHandler;

			m_State = InitStates.None;
			m_Escaped = false;

			m_ResponseBuffer = "";
			m_LastByteWasProcessedCommand = 0;
		}

		#region Public Properties
		/// <summary>
		/// Returns true iff the device is ready for use.
		/// </summary>
		public bool Ready {
			get { return m_State == InitStates.Ready; }
		}

		/// <summary>
		/// The version string returned by the modem.
		/// </summary>
		public string Version { get; private set; }

		/// <summary>
		/// The mode string returned by the modem.
		/// </summary>
		public string Mode { get; private set; }

		/// <summary>
		/// A list of available command strings.
		/// </summary>
		public string[] AvailableCommands { get; private set; }
		#endregion

		#region Pubic Methods
		/// <summary>
		/// Start the communications channel.
		/// </summary>
		public void Start() {
			m_Device.Start();
		}

		/// <summary>
		/// Stop the communications channel.
		/// </summary>
		public void Dispose() {
			m_Device.Dispose();
		}

		/// <summary>
		/// Send an escaped command to the modem.
		/// </summary>
		public bool SendCommand(string command, string argument = null) {
			if (!Ready) return false;

			Debug.WriteLine(String.Format("DEBUG: EndpointSMI.SendCommand({0}, {1})", command, argument ?? "(null)"));

			//
			//  TODO: escape the argument, and check the command for bad characters
			//

			string toSend = null;
			if (String.IsNullOrEmpty(argument)) {
				toSend = String.Format("#{0};", command);
			} else {
				toSend = String.Format("#{0}={1};", command, argument);
			}
			return m_Device.Write(toSend);
		}

		/// <summary>
		/// Send an unescaped string to the port (used internally for initialization)
		/// </summary>
		protected bool SendRaw(string s) {
			Console.WriteLine(String.Format("DEBUG: EndpointSMI.SendRaw(\"{0}\")", s));

			return m_Device.Write(s);
		}

		/// <summary>
		/// Send a stream of data bytes to the modem.
		/// </summary>
		/// <remarks>
		/// Each character in the provided string will be converted into a byte prior to transmission.
		/// </remarks>
		public bool SendData(string data) {
			// don't bother if not ready
			if (!Ready) return false;

			// escape hashes
			data = data.Replace("#", "#HASH;");

			// and send
			return m_Device.Write(data);
		}

		/// <summary>
		/// Send a single data byte to the modem.
		/// </summary>
		/// <remarks>
		/// The character provided will be converted into a byte prior to transmission.
		/// </remarks>
		public bool SendData(char ch) {
			// don't bother if not ready
			if (!Ready) return false;
			
			// escape hashes
			if (ch == '#')
				return m_Device.Write("#HASH;");

			// and send
			return m_Device.Write(ch.ToString());
		}
		#endregion

		#region Events, Handlers and Support Types
		/// <summary>
		/// Describes a command response.
		/// </summary>
		public class CommandResponseEventArgs : EventArgs {
			/// <summary>
			/// The command.
			/// </summary>
			public string Command { get; internal set; }

			/// <summary>
			/// The value.
			/// </summary>
			public string Value { get; internal set; }

			/// <summary>
			/// Success flag.
			/// </summary>
			public bool Success { get; internal set; }
		}

		/// <summary>
		/// Describes received data.
		/// </summary>
		public class DataReceivedEventArgs : EventArgs {
			/// <summary>
			/// Describes a single received data byte, encoded in a char to make string manipulation eaiser.
			/// </summary>
			public char DataByte { get; internal set; }
		}

		/// <summary>
		/// Describes the overall modem state.
		/// </summary>
		public class ModemStateEventArgs : EventArgs {
			/// <summary>
			/// True iff the modem is ready.
			/// </summary>
			public bool Ready { get; internal set; }
		}

		/// <summary>
		/// Raised when a command response is received.
		/// </summary>
		public event EventHandler<CommandResponseEventArgs> CommandResponse;

		/// <summary>
		/// Raised when modem data is received.
		/// </summary>
		public event EventHandler<DataReceivedEventArgs> DataReceived;

		/// <summary>
		/// Raised when the link state changes.
		/// </summary>
		public event EventHandler<LinkStateEventArgs> LinkStateChanged;

		/// <summary>
		/// Raised when the modem state changes.
		/// </summary>
		public event EventHandler<ModemStateEventArgs> ModemStateChanged;

		/// <summary>
		/// Raises and event when a command response received.
		/// </summary>
		protected virtual void OnCommandResponseReceived(object sender, CommandResponseEventArgs e) {
			if (this.CommandResponse != null) {
				this.CommandResponse(sender, e);
			}
		}

		/// <summary>
		/// Raises an event when modem data received.
		/// </summary>
		protected virtual void OnDataReceived(object sender, DataReceivedEventArgs e) {
			if (this.DataReceived != null) {
				this.DataReceived(sender, e);
			}
		}

		/// <summary>
		/// Raises an event when link state changes.
		/// </summary>
		protected virtual void OnLinkStateChanged(object sender, LinkStateEventArgs e) {
			if (this.LinkStateChanged != null) {
				this.LinkStateChanged(sender, e);
			}
		}

		/// <summary>
		/// Raises an event when the modem state changes.
		/// </summary>
		protected virtual void OnModemStateChanged(object sender, ModemStateEventArgs e) {
			if (this.ModemStateChanged != null) {
				this.ModemStateChanged(sender, e);
			}
		}

		/// <summary>
		/// Raised when the link state changes.
		/// </summary>
		private void LinkStateChangeHandler(object sender, LinkStateEventArgs e) {
			// either way, start the sequence over again
			m_State = InitStates.None;
			AvailableCommands = null;
			Version = null;
			Mode = null;

			// send a few #; to wake the device up (TODO: repeat this after a timeout)
			if (e.Connected) {
				SendRaw(@"#;#;#;#;#;");
				CheckInitDelegate d = CheckInit;
				d.BeginInvoke(null, null);
			}

			// signal to owner
			this.OnModemStateChanged(this, new ModemStateEventArgs { Ready = this.Ready });

			// chain to owner
			this.OnLinkStateChanged(this, e);

			// DEBUG:
			Debug.WriteLine("DEBUG: EndpointSMI.LinkStateChanged = {0}", e.Connected);
		}

		/// <summary>
		/// RX Data Handler
		/// </summary>
		private void DeviceDataReceived(object sender, ReceivedDataEventArgs e) {
#if VERBOSE_DEBUG
			Debug.WriteLine(
				String.Format(
					"DEBUG: EndpointSMI.DeviceDataReceived({0}: \"{1}\")",
					e.Data.Length,
					e.Data.Replace("\r", "[CR]").Replace("\n", "[LF]").Replace("/x08", "[BS]")));
#endif
			// loop through each byte in the data packet
			foreach (var ch in e.Data) {
				switch (ch) {
					//
					//  '#'
					//
					case '#': {
						m_Escaped = true;
						m_ResponseBuffer = "";
						m_LastByteWasProcessedCommand = 0;
					} break;

					//
					//  ';'
					//
					case ';': {
						if (m_Escaped) {
							m_Escaped = false;
							ProcessResponseBuffer();
							m_ResponseBuffer = "";
							m_LastByteWasProcessedCommand = 2;
						} else {
							// just a stray ';' so treat it as data
							OnDataReceived(this, new DataReceivedEventArgs { DataByte = ch });
							m_LastByteWasProcessedCommand = 0;
						}
					} break;

					//
					//  CR or LF
					//
					case '\n':
					case '\r': {
						if (m_LastByteWasProcessedCommand != 0) {
							--m_LastByteWasProcessedCommand;
						} else {
							// just a stray CR or LF, so treat it as data
							OnDataReceived(this, new DataReceivedEventArgs { DataByte = ch });
							m_LastByteWasProcessedCommand = 0;
						}
					} break;

					//
					//  all other characters
					//
					default: {
						// ignore unprintable characters
						if (!IsPrint(ch) && (ch != Backspace)) {
							m_LastByteWasProcessedCommand = 0;
							break;
						}

						if (m_Escaped) {
							// if escaped, process incoming command response data
							if (m_ResponseBuffer.Length <= MaximumCommandResponseSize)
								m_ResponseBuffer += ch;
						} else {
							// otherwise, process as RX data
							OnDataReceived(this, new DataReceivedEventArgs { DataByte = ch });
						}
						m_LastByteWasProcessedCommand = 0;
					} break;
				}
			}
		}
		#endregion

		#region Utilities
		/// <summary>
		/// Make sure state is initialized in a reasonable amount of time.
		/// </summary>
		private void CheckInit() {
			Thread.Sleep(m_InitRetryInterval);
			if (m_Device.IsOpen && m_State == InitStates.None) {
				Console.WriteLine("DEBUG: initialization timeout, trying again");
				SendRaw(@"#;#;#;#;#;");
				CheckInitDelegate d = CheckInit;
				d.BeginInvoke(null, null);
			}
		}

		private delegate void CheckInitDelegate();

		/// <summary>
		/// Test whether a character is user-printable.
		/// </summary>
		private static bool IsPrint(char ch) {
			return Char.IsLetterOrDigit(ch) || Char.IsWhiteSpace(ch) || Char.IsSymbol(ch) || Char.IsPunctuation(ch);
		}

		/// <summary>
		/// Process the contents of the command buffer.
		/// </summary>
		private void ProcessResponseBuffer() {
#if VERBOSE_DEBUG
			Debug.WriteLine(String.Format("DEBUG: EndpointSMI.ProcessResponseBuffer({0} bytes)", m_ResponseBuffer.Length));
#endif
			var parts = m_ResponseBuffer.Split(new char [] { '=' }, 2, StringSplitOptions.RemoveEmptyEntries);
			//if ((parts.Length != 2) || parts.Any(f => String.IsNullOrEmpty(f))) return;
			if (parts.Length == 0) return;
			var command = parts[0].Trim();
			var argument = "";
			if (parts.Length >= 2)
				argument = parts[1].Trim();
			ProcessCommandResponse(command, argument);
		}

		/// <summary>
		/// Process a specific command/response pair.
		/// </summary>
		private void ProcessCommandResponse(string command, string argument) {
			Console.WriteLine(String.Format("DEBUG: EndpointSMI.ProcessCommandResponse({0}, {1})", command, argument));

			switch (command.ToUpper()) {
				//
				//  ping response
				//
				case @"OK": {
					if (m_State == InitStates.None) {
						m_State = InitStates.VersionSent;
						// query the device version
						SendRaw(@"#VERSION;");
					}
				} break;

				//
				//  VERSION response
				//
				case @"OK:VERSION": {
					this.Version = argument;
					if (m_State == InitStates.VersionSent) {
						m_State = InitStates.EchoSent;

						// place the modem in echo=2 state, which will echo user data but not commands
						SendRaw(@"#ECHO=2;");
					}

					this.OnCommandResponseReceived(this, new CommandResponseEventArgs { Command = command.Substring(3), Value = argument, Success = true });
				} break;

				//
				//  ECHO response
				//
				case @"OK:ECHO": {
					if (argument == "2") {
						if (m_State == InitStates.EchoSent) {
							m_State = InitStates.ModeSent;

							// query the modem mode
							SendRaw(@"#MODE;");
						}
					} else {
						//
						//  TODO: handle the case when this fails
						//
					}

					this.OnCommandResponseReceived(this, new CommandResponseEventArgs { Command = command.Substring(3), Value = argument, Success = true });
				} break;

				//
				//  MODE response
				//
				case @"OK:MODE": {
					Mode = argument;
					if (m_State == InitStates.ModeSent) {
						m_State = InitStates.DocSent;

						// query the doc string
						SendRaw(@"#DOC;");
					}

					// signal the owner
					this.OnModemStateChanged(this, new ModemStateEventArgs { Ready = this.Ready });
					this.OnCommandResponseReceived(this, new CommandResponseEventArgs { Command = command.Substring(3), Value = argument, Success = true });
				} break;

				//
				//  DOC response
				//
				case @"OK:DOC": {
					AvailableCommands = argument.Split(',');
					if (m_State == InitStates.DocSent) {
						m_State = InitStates.Ready;
					}

					this.OnCommandResponseReceived(this, new CommandResponseEventArgs { Command = command.Substring(3), Value = argument, Success = true });
				} break;

				//
				//  HASH (#) escape code
				//
				case @"HASH": {
					this.OnDataReceived(this, new DataReceivedEventArgs { DataByte = '#' });
				} break;

				//
				//  all other command responses
				//
				default: {
					if (command.StartsWith(@"OK:")) {
						this.OnCommandResponseReceived(this, new CommandResponseEventArgs { Command = command.Substring(3), Value = argument, Success = true });
					} else if (command.StartsWith(@"ERR:")) {
						this.OnCommandResponseReceived(this, new CommandResponseEventArgs { Command = command.Substring(4), Value = argument, Success = false });
					} else {
						this.OnCommandResponseReceived(this, new CommandResponseEventArgs { Command = command, Value = argument, Success = true });
					}
				} break;
			}
		}
		#endregion
	}
}
