﻿/*
 *
 *
 *    ConfigDialog.cs
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2017 by Matthew K. Roberts, KK5JY. All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *    
 *
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using KK5JY.Config;
using Gtk;
using KK5JY.IO;

namespace KK5JY.SMI.GTK {
	/// <summary>
	/// A dialog for changing application settings (serial port & speed, callsign, etc).
	/// </summary>
	public sealed class ConfigDialog : Dialog {
		#region Constants
		/// <summary>
		/// Valid serial port speeds.
		/// </summary>
		public static readonly string[] Speeds = { "115200", "57600", "38400", "19200", "9600", "4800", "1200" };

		/// <summary>
		/// Valid serial port names.
		/// </summary>
		public readonly string[] Ports;
		#endregion

		#region Private Fields
		private Config m_ConfigCache;

		private Entry eCall;
		private ComboBoxEntry cbPort;
		private ComboBox cbSpeed;
		private Entry eLogger;
		private FontButton bFont;
		private ColorButton bForeground;
		private ColorButton bBackground;
		private CheckButton cbDupe;
		private CheckButton cbSelect;
		#endregion

		#region Constructor
		/// <summary>
		/// Create a new <c>ConfigDialog</c>.
		/// </summary>
		/// <param name="parent">The parent/owning window.</param>
		public ConfigDialog(Window parent) : base("SMI Configuration", parent, Gtk.DialogFlags.DestroyWithParent) {
			// configure the window itself
			this.Title = "SMI Configuration";
			this.Deletable = true;
			this.Resizable = false;
			this.Modal = true;

			var table = new Table(8, 3, false);
			this.VBox.PackStart(table, false, true, 0);
			uint row = 0;

			var lCall = new Label("My Call:");
			eCall = new Entry();
			table.Attach(lCall, 0, 1, row, row + 1);
			table.Attach(eCall, 1, 3, row, row + 1);
			++row;

			Ports = SerialPortHelper.GetPortNames();
			var lPort = new Label("Modem Port:");
			cbPort = new ComboBoxEntry(Ports);
			table.Attach(lPort, 0, 1, row, row + 1);
			table.Attach(cbPort, 1, 3, row, row + 1);
			++row;

			var lSpeed = new Label("Modem Speed:");
			cbSpeed = new ComboBox(Speeds);
			table.Attach(lSpeed, 0, 1, row, row + 1);
			table.Attach(cbSpeed, 1, 3, row, row + 1);
			++row;

			var lLog = new Label("Log Editor Path:");
			var bLogger = new Button { Label = "..." };
			bLogger.Clicked += new EventHandler(LoggerPathClick);
			eLogger = new Entry();
			table.Attach(lLog, 0, 1, row, row + 1);
			table.Attach(eLogger, 1, 2, row, row + 1);
			table.Attach(bLogger, 2, 3, row, row + 1);
			++row;

			cbDupe = new CheckButton("Dupe checking includes band info");
			table.Attach(cbDupe, 0, 3, row, row + 1);
			++row;

			cbSelect = new CheckButton("Select text when entering log field");
			table.Attach(cbSelect, 0, 3, row, row + 1);
			++row;

			var lFont = new Label("QSO Font:");
			bFont = new FontButton();
			table.Attach(lFont, 0, 1, row, row + 1);
			table.Attach(bFont, 1, 3, row, row + 1);
			++row;

			var lFore = new Label("QSO Foreground:");
			bForeground = new ColorButton();
			table.Attach(lFore, 0, 1, row, row + 1);
			table.Attach(bForeground, 1, 3, row, row + 1);
			++row;

			var lBack = new Label("QSO Background:");
			bBackground = new ColorButton();
			table.Attach(lBack, 0, 1, row, row + 1);
			table.Attach(bBackground, 1, 3, row, row + 1);
			++row;

			foreach (var label in new[] { lCall, lPort, lSpeed, lLog, lFont, lFore, lBack }) {
				label.SetAlignment(1.0f, 0.5f);
				label.SetPadding(7, 0);
			}

			this.AddButton("OK", ResponseType.Ok);
			this.AddButton("Cancel", ResponseType.Cancel);
			this.SetResponseSensitive(ResponseType.Ok, true);
			this.SetResponseSensitive(ResponseType.Cancel, true);
			this.DefaultResponse = ResponseType.Ok;

			ShowAll();
		}
		#endregion

		#region Public Methods
		/// <summary>
		/// Get the system configuration.
		/// </summary>
		public Config GetConfiguration() {
			m_ConfigCache.MyCall = eCall.Text;
			m_ConfigCache.Speed = UInt32.Parse(cbSpeed.ActiveText);
			m_ConfigCache.Port = cbPort.ActiveText;
			m_ConfigCache.LoggerPath = eLogger.Text;

			m_ConfigCache.WindowBackground.FromColor(bBackground.Color);
			m_ConfigCache.WindowForeground.FromColor(bForeground.Color);
			m_ConfigCache.WindowFont.FromFont(Pango.FontDescription.FromString(bFont.FontName));

			m_ConfigCache.DupeCheckIncludesBand = cbDupe.Active;
			m_ConfigCache.SelectLogTextOnEntry = cbSelect.Active;

			return m_ConfigCache;
		}

		/// <summary>
		/// Set the system configuration.
		/// </summary>
		public void SetConfiguration(Config value) {
			// core config
			m_ConfigCache = value;
			eCall.Text = value.MyCall ?? "";
			int idx = -1;
			if (!String.IsNullOrEmpty(value.Port)) {
				idx = Array.IndexOf(Ports, value.Port);
				if (idx >= 0)
					cbPort.Active = idx;
			}
			idx = Array.IndexOf(Speeds, value.Speed.ToString());
			if (idx >= 0)
				cbSpeed.Active = idx;

			// GUI config
			if (!String.IsNullOrEmpty(m_ConfigCache.LoggerPath)) {
				eLogger.Text = m_ConfigCache.LoggerPath;
			}
			if (!ReferenceEquals(value.WindowForeground, null)) {
				bForeground.Color = new Gdk.Color(
					Convert.ToByte(value.WindowForeground.R / 256),
					Convert.ToByte(value.WindowForeground.G / 256),
					Convert.ToByte(value.WindowForeground.B / 256)
				);
			}
			if (!ReferenceEquals(value.WindowBackground, null)) {
				bBackground.Color = new Gdk.Color(
					Convert.ToByte(value.WindowBackground.R / 256),
					Convert.ToByte(value.WindowBackground.G / 256),
					Convert.ToByte(value.WindowBackground.B / 256)
				);
			}
			if (!ReferenceEquals(value.WindowFont, null)) {
				bFont.SetFontName(String.Format("{0} {1} {2}",
					value.WindowFont.FontFamily,
					value.WindowFont.Style,
					value.WindowFont.Size
				));
			}
			bFont.ShowSize = true;
			bFont.UseSize = true;

			cbDupe.Active = value.DupeCheckIncludesBand;
			cbSelect.Active = value.SelectLogTextOnEntry;
		}
		#endregion

		#region UI Event Handlers
		/// <summary>
		/// Raised when the logger path elipsis (...) is clicked.
		/// </summary>
		private void LoggerPathClick(object sender, EventArgs e) {
			using (var dialog = new FileChooserDialog("Choose Logging Program", this, FileChooserAction.Open, "OK", ResponseType.Ok, "Cancel", ResponseType.Cancel)) {
				// add filters
				FileFilter filter = null;
				if (ConfigUtils.IsWindows) {
					filter = new FileFilter { Name = "Executable Files (*.exe)" };
					filter.AddMimeType("text/plain");
					filter.AddPattern("*.txt");
					dialog.AddFilter(filter);
				}
				filter = new FileFilter { Name = "All Files (*.*)" };
				filter.AddPattern("*.*");
				dialog.AddFilter(filter);

				// set initial path
				if (!String.IsNullOrEmpty(m_ConfigCache.LoggerPath)) {
					dialog.SetFilename(m_ConfigCache.LoggerPath);
				}

				// run the dialog
				var result = (ResponseType)(dialog.Run());
				if (result == ResponseType.Ok) {
					eLogger.Text = dialog.Filename;
				}
				dialog.Destroy();
			}
		}
		#endregion
	}
}
