﻿/*
 *
 *
 *    MainWindow.cs
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2017 by Matthew K. Roberts, KK5JY. All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *    
 *
 */

// define this symbol to generate fake noisy data to test UI logic related
//    to handling incoming data (TESTING ONLY)
//#define FAKE_RECEIVER_DATA

// define this to cause the font styles for the log entry boxes to follow
//    the styles set by the user for the QSO window
//#define LOG_ENTRIES_FOLLOW_WINDOW_STYLE

using System;
using System.Collections.Generic;
using System.Collections.Concurrent;
using System.Globalization;
using System.Xml.Serialization;
using System.Linq;
using System.Text;
using System.IO;
using System.Threading;
using System.Diagnostics;
using System.IO.Ports;

using Gtk;

using KK5JY.Config;
using KK5JY.Log;
using KK5JY.XMLRPC;
using KK5JY.IO;

namespace KK5JY.SMI.GTK {
	/// <summary>
	/// Main window type for GTK front-end.
	/// </summary>
	public sealed partial class MainWindow : Window {
		#region UI Elements
		private const string BreakInOn = "Break-In";
		private const string BreakInOff = "TX Off";
		private const string AppTitle = "SMI Console";
		private const int MainWindowMargin = 4;
		private const int MaxWindowLength = 64 * 1024; // max 64k of QSO scrollback
		private const int WindowLengthDecrement = 128; // remove this many bytes when above limit exceeded
		private const int MacroButtonCount = 12;

		private FilteredEntry tb_Call, tb_RX;
		private Entry tb_TX, tb_Notes;
		private Label l_Call, l_RX, l_TX, l_Notes, l_Freq;
		private ComboBoxEntry cb_Freq;
		private ToggleButton cb_BreakIn;
		private Button b_Log, b_Clear;
		private TerminalWindow tb_Main;
		private ScrolledWindow m_Scroller;
		private Statusbar sb_Status;
		private Button[] MacroButtons;
		private HBox m_ToolbarLayout;
		private CheckMenuItem fs_item;

		private void InitializeWindow() {
			// configure the window itself
			this.Title = AppTitle;
			this.Deletable = true;
			this.Resize(600, 400);

			string iconFile = "SMI.ico";
			iconFile = System.IO.Path.Combine(
				System.IO.Path.GetDirectoryName(System.Reflection.Assembly.GetExecutingAssembly().Location),
				iconFile);
			if (File.Exists(iconFile)) {
				try {
					this.SetIconFromFile(iconFile);
				} catch (Exception ex) {
					Console.WriteLine("Could not load icon: {0}", ex.Message);
				}
			} else {
				Console.WriteLine("Could not find icon: {0}", iconFile);
			}

			// follow window state changes
			this.WindowStateEvent += WindowStateChangeEvent;

			// menu
			MenuItem file_item = new MenuItem("File");
			Menu file_menu = new Menu();
			file_item.Submenu = file_menu;

			MenuItem new_item = new MenuItem("Save QSO Text As...");
			new_item.Activated += new EventHandler(SaveAsClick);
			file_menu.Append(new_item);

			MenuItem capture_item = new MenuItem("Capture QSO Text...");
			capture_item.Activated += new EventHandler(CaptureClick);
			file_menu.Append(capture_item);

			MenuItem close_capture_item = new MenuItem("Close QSO Capture");
			close_capture_item.Activated += new EventHandler(CloseCaptureClick);
			file_menu.Append(close_capture_item);

			file_menu.Append(new SeparatorMenuItem());

			MenuItem settings_item = new MenuItem("Configure SMI...");
			settings_item.Activated += new EventHandler(ConfigClick);
			file_menu.Append(settings_item);

			MenuItem savenow_item = new MenuItem("Save Current Configuration");
			savenow_item.Activated += new EventHandler(SaveConfigNowClick);
			file_menu.Append(savenow_item);

			settings_item = new MenuItem("Modem Settings...");
			settings_item.Activated += new EventHandler(ModemSettingsClick);
			file_menu.Append(settings_item);

			file_menu.Append(new SeparatorMenuItem());

			MenuItem exit_item = new MenuItem("Exit");
			exit_item.Activated += new EventHandler(ExitClick);
			file_menu.Append(exit_item);

			MenuItem view_item = new MenuItem("View");
			Menu view_menu = new Menu();
			view_item.Submenu = view_menu;

			fs_item = new CheckMenuItem("Full Screen");
			fs_item.Activated += new EventHandler(FullScreenClick);
			view_menu.Append(fs_item);

			MenuItem log_item = new MenuItem("Log");
			Menu log_menu = new Menu();
			log_item.Submenu = log_menu;

			new_item = new MenuItem("View Log...");
			new_item.Activated += new EventHandler(LogViewClick);
			log_menu.Append(new_item);

			new_item = new MenuItem("Select File...");
			new_item.Activated += new EventHandler(SelectLogFileClick);
			log_menu.Append(new_item);

			MenuItem help_item = new MenuItem("Help");
			help_item.RightJustified = true;
			Menu help_menu = new Menu();
			help_item.Submenu = help_menu;

			new_item = new MenuItem("About");
			new_item.Activated += new EventHandler(AboutClick);
			help_menu.Append(new_item);

			MenuBar mb = new MenuBar();
			mb.Append(file_item);
			mb.Append(view_item);
			mb.Append(log_item);
			mb.Append(help_item);

			// what are the GTK guys smoking?  There's no other way to turn off F10 for "show menu"???
			// Ref: https://groups.google.com/forum/#!topic/wx-users/mSQTZFV4T5c
			Gtk.Settings.Default.SetStringProperty("gtk-menu-bar-accel", "F15", "foo");

			// layout containers
			var rootLayout = new VBox();
			var logLayout = new Table(2, 8, false);
			var mainLayout = new HBox();
			m_ToolbarLayout = new HBox();
			var mainSplitter = new HPaned();
			sb_Status = new Statusbar();
			mainLayout.Add(mainSplitter);
			rootLayout.PackStart(mb, false, true, 0);
			rootLayout.PackStart(logLayout, false, true, 0);
			rootLayout.PackStart(new HSeparator(), false, false, 2);
			rootLayout.PackStart(mainLayout, true, true, 0);
			rootLayout.PackStart(m_ToolbarLayout, false, false, 0);
			rootLayout.PackStart(sb_Status, false, true, 0);
			this.Add(rootLayout);

			// first row
			l_Call = new Label("Call:");
			l_Call.SetPadding(5, 0);
			tb_Call = new FilteredEntry();
			tb_Call.InsertRequest += CallTextInsertRequest;
			tb_Call.Changed += CallTextChanged;
			tb_Call.FocusGrabbed += LogFieldEntered;
			l_RX = new Label("RX:");
			l_RX.SetPadding(5, 0);
			tb_RX = new FilteredEntry();
			tb_RX.InsertRequest += RxTextInsertRequest;
			tb_RX.FocusGrabbed += LogFieldEntered;
			l_TX = new Label("TX:");
			l_TX.SetPadding(5, 0);
			tb_TX = new Entry();
			tb_TX.FocusGrabbed += LogFieldEntered;
			b_Log = new Button("Log");
			b_Log.Pressed += new EventHandler(LogClicked);
			b_Clear = new Button("Clear");
			b_Clear.Pressed += new EventHandler(ClearClicked);

			// second row
			l_Freq = new Label("Freq:");
			l_Freq.SetPadding(5, 0);
			cb_Freq = new ComboBoxEntry(Bands);
			cb_Freq.Active = 1;
			cb_Freq.Changed += new EventHandler(cbFreq_Changed);
			l_Notes = new Label("Notes:");
			l_Notes.SetPadding(5, 0);
			tb_Notes = new Entry();
			tb_Notes.FocusGrabbed += LogFieldEntered;
			cb_BreakIn = new ToggleButton();
			cb_BreakIn.Label = BreakInOff;
			cb_BreakIn.Toggled += new EventHandler(BreakInClicked);

			// first row
			logLayout.Attach(l_Call, 0, 1, 0, 1, AttachOptions.Fill, AttachOptions.Fill, 3, 1);
			logLayout.Attach(tb_Call, 1, 2, 0, 1);
			logLayout.Attach(l_RX, 2, 3, 0, 1, AttachOptions.Fill, AttachOptions.Fill, 3, 1);
			logLayout.Attach(tb_RX, 3, 4, 0, 1);
			logLayout.Attach(l_TX, 4, 5, 0, 1, AttachOptions.Fill, AttachOptions.Fill, 3, 1);
			logLayout.Attach(tb_TX, 5, 6, 0, 1);
			logLayout.Attach(b_Log, 6, 7, 0, 1, AttachOptions.Shrink, AttachOptions.Fill, 3, 1);
			logLayout.Attach(b_Clear, 7, 8, 0, 1, AttachOptions.Shrink, AttachOptions.Fill, 3, 1);

			// second row
			logLayout.Attach(l_Freq, 0, 1, 1, 2, AttachOptions.Fill, AttachOptions.Fill, 3, 1);
			logLayout.Attach(cb_Freq, 1, 2, 1, 2);
			logLayout.Attach(l_Notes, 2, 3, 1, 2, AttachOptions.Fill, AttachOptions.Fill, 3, 1);
			logLayout.Attach(tb_Notes, 3, 6, 1, 2);
			logLayout.Attach(cb_BreakIn, 6, 8, 1, 2, AttachOptions.Fill, AttachOptions.Fill, 3, 1);

			// main QSO window
			tb_Main = new TerminalWindow();
			tb_Main.LeftMargin += MainWindowMargin;
			tb_Main.UnfilteredKeyPress += QsoWindowKeyPress;
			tb_Main.PopulatePopup += MainWindowPopulatePopup;
			m_Scroller = new ScrolledWindow();
			m_Scroller.Add(tb_Main);
			mainSplitter.Pack1(m_Scroller, true, false);
			var macroBar = new VBox();
			this.KeyPressEvent += MainWindowKeyPressEvent;
			this.KeyReleaseEvent += MainWindowKeyReleaseEvent;

			// macro bar
			mainSplitter.Pack2(macroBar, false, false);
			MacroButtons = new Button[MacroButtonCount];
			for (int i = 0; i != MacroButtons.Length; ++i) {
				var b = new Button(String.Format("F{0}", i + 1));
				b.SetAlignment(0.0f, 0.5f);
				b.Pressed += MacroButtonPressed;
				MacroButtons[i] = b;
				macroBar.Add(b);
			}

			//
			//  touch up some items
			//

			// create a new font descriptor that bolds the default label font
			var newLabelFont = l_Call.Style.FontDesc.Copy();
			newLabelFont.Weight = Pango.Weight.Bold;
			foreach (var widget in new [] { l_Call, l_RX, l_TX, l_Freq, l_Notes }) {
				widget.ModifyFont(newLabelFont);
				widget.SetAlignment(0.9f, 0.5f);
			}

			// do the same for some of the text boxes
			var newEntryFont = tb_Call.Style.FontDesc.Copy();
			newEntryFont.Weight = Pango.Weight.Bold;
			foreach (var widget in new Widget[] { tb_Call, tb_RX, tb_TX, cb_Freq, cb_BreakIn.Child as Label }) {
				widget.ModifyFont(newEntryFont);
			}

			// and for some of the buttons
			var newButtonFont = b_Log.Style.FontDesc.Copy();
			newButtonFont.Weight = Pango.Weight.Bold;
			foreach (var button in new Button[] { b_Log }) {
				button.Child.ModifyFont(newButtonFont);
			}
			foreach (var button in MacroButtons) {
				button.Child.ModifyFont(newButtonFont);
			}

			//
			//  wire up the special navigation handler
			//
			foreach (var widget in new Widget[] { tb_Call, tb_RX, tb_TX, tb_Notes, cb_Freq, b_Log }) {
				widget.KeyPressEvent += SpecialNavigationKeyHandler;
			}
			foreach (var button in MacroButtons) {
				button.KeyPressEvent += SpecialNavigationKeyHandler;
			}

			//
			//  configure tab stops
			//
			this.FocusChain = new Widget[] { logLayout, mainLayout };
			logLayout.FocusChain = new Widget[] { tb_Call, tb_RX };

			//
			//  auto-scrolling
			//
			tb_Main.SizeRequested += new SizeRequestedHandler(QsoWindowSizeRequested);
		}

		/// <summary>
		/// Called when a log field is entered.
		/// </summary>
		private void LogFieldEntered(object o, EventArgs args) {
			Console.WriteLine("LogFieldEntered()");

			var entry = o as Entry;
			if (ReferenceEquals(entry, null)) {
				Console.WriteLine("WARNING: LogFieldEntered called on non-Entry control");
				return;
			}

			// de-select the text in the box
			if ( ! m_Config.SelectLogTextOnEntry) {
				int pos = entry.Text.Length;
				entry.SelectRegion(pos, pos);
			}
		}
		#endregion

		#region Test Objects
		#if FAKE_RECEIVER_DATA
		/// <summary>
		/// A timer for periodically appending text to the QSO window.
		/// </summary>
		private System.Threading.Timer m_TestTimer;

		/// <summary>
		/// Mix in some test data to simulate asynchronous modem data.
		/// </summary>
		private void TestTimerTick(object state) {
			Console.WriteLine("DEBUG: TestTimerDelegate()");
			const string TestDataSet = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
			var prng = new Random(Convert.ToInt32(DateTime.UtcNow.Ticks & 0x7FFFFFFF));
			var sb = new StringBuilder();
			for (int i = 0; i != 25; ++i) {
				for (int j = 0; j != 5; ++j) {
					sb.Append(TestDataSet[prng.Next(TestDataSet.Length)]);
				}
				sb.Append(" ");
			}
			sb.Append("\n");

			SafeAppendText(sb.ToString());
		}
		#endif
		#endregion

		#region Constants
		/// <summary>
		/// Escape character.
		/// </summary>
		const char ESC = '\x1B';

		/// <summary>
		/// Backspace character.
		/// </summary>
		const char BACKSPACE = '\x08';

		public const string AppName = @"SMI-GTK";
		public const string AppAuthor = @"KK5JY.Net";

		public static readonly string ConfigRoot =
			System.Environment.GetFolderPath(
				Environment.SpecialFolder.ApplicationData
			);
		public static readonly string ConfigPath = System.IO.Path.Combine(
			ConfigRoot,
			AppAuthor,
			AppName);
		public static readonly string ConfigFile = System.IO.Path.Combine(
			ConfigRoot,
			AppAuthor,
			AppName,
			@"Config.xml");
		public static readonly string DefaultLogFile = System.IO.Path.Combine(
			ConfigRoot,
			AppAuthor,
			AppName,
			@"Logbook.adi");

		// list of bands for the combobox -- the first one is blank, to allow an
		//   index into which to place automatic readings from flrig
		public static readonly string[] Bands = { "", "160m", "80m", "60m", "40m", "30m", "20m", "17m", "15m", "12m", "10m", "6m", "2m" };

		// the about box text
		public const string AboutText = @"SMI Console, Version 1.0 (beta)

Copyright (C) 2017 by Matt Roberts, KK5JY,
All rights reserved.

License: GNU General Public License Version 3.0.
    
This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
    
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
    
You should have received a copy of the GNU General Public License along with this program.  If not, see:

http://www.gnu.org/licenses/";

		/// <summary>
		/// How often to poll for frequency.
		/// </summary>
		private readonly TimeSpan FreqInterval = TimeSpan.FromSeconds(1);

		/// <summary>
		/// The longest time between log count queries.
		/// </summary>
		private readonly TimeSpan LogInterval = TimeSpan.FromSeconds(5);
		#endregion

		#region Private Fields
		/// <summary>
		/// The current window state.
		/// </summary>
		public Gdk.WindowState WindowState { get; private set; }

		/// <summary>
		/// UI update timer.
		/// </summary>
		private Timer m_UpdateTimer;

		/// <summary>
		/// The capture file.
		/// </summary>
		private FileStream m_CaptureStream;

		/// <summary>
		/// The last time a timestamp was written to the capture file.
		/// </summary>
		private DateTime m_CaptureStamp;

		/// <summary>
		/// Stores the time the call was entered as "time on".
		/// </summary>
		private DateTime m_LastTimeOn;

		/// <summary>
		/// Tracks CR state for text being cleaned for the main window.
		/// </summary>
		private bool m_LastWasCR;

		/// <summary>
		/// The configuration.
		/// </summary>
		private Config m_Config;

		/// <summary>
		/// The modem.
		/// </summary>
		private EndpointSMI m_Modem;

		/// <summary>
		/// The macro expander.
		/// </summary>
		private MacroExpansion m_MacroExpander;

		/// <summary>
		/// The log file.
		/// </summary>
		private AdifFile m_Log;

		/// <summary>
		/// The rig connection.
		/// </summary>
		private FlRig m_Rig;

		/// <summary>
		/// Return the current local break-in state.
		/// </summary>
		private BreakIn BreakInState {
			get { return cb_BreakIn.Active ? BreakIn.Enabled : BreakIn.Disabled; }
		}

		/// <summary>
		/// Returns the current modem mode.
		/// </summary>
		private string Mode {
			get { return "No Mode"; }
		}

		/// <summary>
		/// Time of the last frequency query.
		/// </summary>
		private DateTime LastFrequencyQuery;

		/// <summary>
		/// Control state, used to eliminate unnecessary UI repaints.
		/// </summary>
		ushort m_BackspaceState;

		/// <summary>
		/// The output buffer for the main text box.
		/// </summary>
		private ConcurrentQueue<char> m_WindowBuffer;

		/// <summary>
		/// Track CTRL key use.
		/// </summary>
		private bool m_CtrlDown = false;

		/// <summary>
		/// Cache for settings values.
		/// </summary>
		private ConcurrentDictionary<string, string> m_SettingsValues;

		/// <summary>
		/// The normal color for settings base.
		/// </summary>
		private Gdk.Color m_NormalSettingBase;

		/// <summary>
		/// Outstanding settings requests.
		/// </summary>
		private LinkedList<string> m_OutstandingQueries;

		/// <summary>
		/// The cached count of log entries.
		/// </summary>
		private int m_LogCount;

		/// <summary>
		/// The time of the last log count query.
		/// </summary>
		private DateTime m_LogQuery;
		#endregion

		#region Inner Types
		private delegate string GetStringDelegate();
		private delegate decimal GetDecimalDelegate();
		private delegate void SetFreqDelegate(decimal d);
		private delegate void SetDupeDelegate(bool b);
		private delegate void CheckDupeDelegate(string c, int b, string m);
		private delegate void RebuildSettingsBarDelegate();
		private delegate void BackspaceDelegate();
		private delegate void AppendTextDelegate();

		private class SetFrequencyArgs : EventArgs {
			public decimal Frequency { get; set; }
		}
		#endregion

		#region Static Fields
		/// <summary>
		/// Serializer used to save/restore state/config.
		/// </summary>
		static XmlSerializer s_ConfigSerializer;
		#endregion

		#region Constructors
		/// <summary>
		/// Static ctor.
		/// </summary>
		static MainWindow() {
			s_ConfigSerializer = new XmlSerializer(typeof(Config));
		}

		/// <summary>
		/// Default ctor.
		/// </summary>
		public MainWindow() : base(WindowType.Toplevel) {
			// initialize and layout the UI components
			InitializeWindow();

			// construct a default configuration, and read in some style info from default widgets
			m_Config = new Config();
			m_Config.WindowFont = new SerializableFont();
			m_Config.WindowFont.FromFont(tb_Main.Font);
			m_Config.WindowForeground = new SerializableColor();
			m_Config.WindowForeground.FromColor(tb_Main.ForeColor);
			m_Config.WindowBackground = new SerializableColor();
			m_Config.WindowBackground.FromColor(tb_Main.BackColor);
			m_Config.GoodColor = new SerializableColor { R = 0, G = 65535, B = 0 }; // green
			m_Config.WarnColor = new SerializableColor { R = 65535, G = 65535, B = 0 }; // yellow

			// initialize the macro expander
			m_MacroExpander = new MacroExpansion();
			m_MacroExpander.FieldRequest += MacroFieldRequest;

			// the XMLRPC interface
			m_Rig = new FlRig();

			// the default log
			m_Log = new AdifFile(DefaultLogFile);

			// main window buffer
			m_WindowBuffer = new ConcurrentQueue<char>();

			// other stuff
			m_SettingsValues = new ConcurrentDictionary<string, string>(StringComparer.InvariantCultureIgnoreCase);
			m_LogCount = -1;
			m_LogQuery = DateTime.MinValue;

			#if FAKE_RECEIVER_DATA
			m_TestTimer = new System.Threading.Timer(TestTimerTick, null, 3000, 2000);
			#endif
		}
		#endregion

		#region Utilities
		/// <summary>
		/// Reads the current band in meters.
		/// </summary>
		private int GetBand() {
			string working = cb_Freq.ActiveText.ToLower().Trim();

			// if the display is already in meters, just use that
			if (working.EndsWith("m")) {
				working = working.Substring(0, working.Length - 1);
				return Int32.Parse(working);
			}

			// if the frequency is in kHz, use the ADIF tools to convert it
			decimal freq;
			if (Decimal.TryParse(working, out freq)) {
				return AdifRecord.GetBand(freq / 1000m);
			}

			return 0;
		}

		/// <summary>
		/// Check for dupe.
		/// </summary>
		private void CheckDupe(string call, int ? bandInMeters, string mode) {
			bool isDupe = false;
			if (!String.IsNullOrEmpty(call)) {
				// use the log search to determine if this is a dupe
				isDupe = m_Log.GetRecords(call.ToUpper(), bandInMeters, mode).FirstOrDefault() != null;
			}

			// and update the UI accordingly
			this.SetDupe(isDupe);

			// DEBUG:
			Debug.WriteLine("DEBUG: CheckDupe({0}, {1}, {2}) = {3}", call, bandInMeters, mode, isDupe);
		}

		/// <summary>
		/// Perform duplicate-contact checking.
		/// </summary>
		private void CheckDupe() {
			// get a reliable reference to the modem
			var modem = m_Modem;
			if (ReferenceEquals(modem, null))
				return;

			// see if the current call is a duplicate
			CheckDupe(tb_Call.Text, m_Config.DupeCheckIncludesBand ? (int?)GetBand() : null, modem.Mode);
		}

		/// <summary>
		/// Test whether a character is user-printable.
		/// </summary>
		private static bool IsPrint(char ch) {
			return Char.IsLetterOrDigit(ch) || Char.IsWhiteSpace(ch) || Char.IsSymbol(ch) || Char.IsPunctuation(ch);
		}

		/// <summary>
		/// Send data to modem (single character).
		/// </summary>
		/// <param name="ch">The character to send.</param>
		/// <returns>True iff data sent successfully.</returns>
		private bool SendData(char ch) {
			// if tx disabled, bail
			if (BreakInState == BreakIn.Disabled) {
				return false;
			}

			// no modem? quit now
			if (ReferenceEquals(m_Modem, null)) {
				return false;
			}

			// send
			return m_Modem.SendData(ch);
		}

		/// <summary>
		/// Send data to the modem (string).
		/// </summary>
		/// <param name="s">The text to send.</param>
		/// <returns>True iff all data sent successfully.</returns>
		private bool SendModemData(string s) {
			// dont' bother if bk-in disabled
			if (BreakInState != BreakIn.Enabled) return false;

			bool status = true;
			foreach (char ch in s) {
				status &= SendData(ch);
				if (!status) break;
			}
			if (!status) {
				Debug.WriteLine("DEBUG: Error - not all macro bytes were sent");
			}
			return status;
		}

		/// <summary>
		/// Apply updated configuration information.  Restarts connection to modem with new port setup.
		/// </summary>
		private void RestartModem() {
			if (!ReferenceEquals(m_Modem, null)) {
				m_Modem.Dispose();
			}

			if (String.IsNullOrEmpty(m_Config.Port))
				return;
			if (!SerialPortHelper.GetPortNames().Contains(m_Config.Port)) {
				using (var dialog = new MessageDialog(
						this, DialogFlags.DestroyWithParent, MessageType.Warning, ButtonsType.Ok,
						"The specified serial port name is invalid.  Please check the configuration.") { Title = "Invalid Modem Port" }) {
					dialog.Run();
					dialog.Destroy();
				}
				return;
			}

			m_Modem =
				new EndpointSMI(
					new SerialPortHelper(
						new SerialPort(
							m_Config.Port,
							Convert.ToInt32(m_Config.Speed))));
			m_Modem.CommandResponse += ModemCommandResponse;
			m_Modem.DataReceived += ModemDataReceived;
			m_Modem.LinkStateChanged += ModemLinkStateChanged;
			m_Modem.ModemStateChanged += ModemOverallStateChanged;
			m_Modem.Start();
		}

		/// <summary>
		/// Clean 'noise' characters out of a call.
		/// </summary>
		private static string FilterCall(string call) {
			var sb = new StringBuilder();
			foreach (char ch in call) {
				if (Char.IsWhiteSpace(ch)) {
					continue;
				}
				if (Char.IsPunctuation(ch) && (ch != '/') && (ch != '-')) {
					continue;
				}
				sb.Append(ch);
			}
			return sb.ToString();
		}

		/// <summary>
		/// Clean 'noise' characters out of an exchange.
		/// </summary>
		private static string FilterExchange(string call) {
			if (ReferenceEquals(call, null)) return null;
			return call.Replace("\n", "").Replace("\r", "").Replace("\t", " ");
		}

		/// <summary>
		/// Thunk between timer and UI.
		/// </summary>
		private void TimerCallback(object state) {
			Application.Invoke(UpdateTimerTick);
		}

		/// <summary>
		/// Return the number of log entries.
		/// </summary>
		private int LogCount() {
			try {
				return m_Log.Count;
			} catch {
				return 0;
			}
		}

		#region QSO Window Utilities
		/// <summary>
		/// Make consistent line endings.
		/// </summary>
		private string ConvertEndings(string s) {
			var sb = new StringBuilder();
			foreach (char ch in s) {
				switch (ch) {
					case '\r': {
						m_LastWasCR = true;
					} break;
					case '\n': {
						if (!m_LastWasCR)
							sb.Append(Environment.NewLine);
						else
							sb.Append(ch);
						m_LastWasCR = false;
					} break;
					default: {
						m_LastWasCR = false;
						sb.Append(ch);
					} break;
				}
			}
			return sb.ToString();
		}

		/// <summary>
		/// Append a string to the text.
		/// </summary>
		private void SafeAppendText(string s) {
			foreach (var ch in s)
				m_WindowBuffer.Enqueue(ch);
			EventHandler d = SafeAppendTextWorker;
			Application.Invoke(d);
		}

		/// <summary>
		/// Append a character to the text.
		/// </summary>
		private void SafeAppendText(char ch) {
			m_WindowBuffer.Enqueue(ch);
			EventHandler d = SafeAppendTextWorker;
			Application.Invoke(d);
		}

		/// <summary>
		/// Invocation target.
		/// </summary>
		private void SafeAppendTextWorker(object sender, EventArgs args) {
			var sb = new StringBuilder();
			char ch;
			while (m_WindowBuffer.TryDequeue(out ch)) {
				sb.Append(ch);
			}
			this.AppendTextWorker(sb.ToString());
		}

		/// <summary>
		/// Insert text at the end of the text window, without disturbing any
		/// ongoing selection operations.
		/// </summary>
		private void AppendTextWorker(string s) {
			int idx = 0;
			while ((idx = s.IndexOf(BACKSPACE)) >= 0) {
				if (idx == 0) {
					// delete one character from the window
					if (tb_Main.Buffer.CharCount != 0) {
						var where = tb_Main.Buffer.GetIterAtOffset(tb_Main.Buffer.CharCount - 1);
						tb_Main.Buffer.Backspace(ref where, false, false);
					}
				} else {
					s = s.Substring(0, idx) + s.Substring(idx + 1);
				}
			}

			// make sure we have consistent line endings
			s = ConvertEndings(s);

			//
			//  using this sequence appears to leave the selection intact
			//
			var iter = tb_Main.Buffer.EndIter;
			tb_Main.Buffer.Insert(ref iter, s);
			if (tb_Main.Buffer.CharCount >= MaxWindowLength) {
				var iter1 = tb_Main.Buffer.GetIterAtOffset(0);
				var iter2 = tb_Main.Buffer.GetIterAtOffset(WindowLengthDecrement);
				Console.WriteLine("Auto-trim {0} characters from QSO window.", WindowLengthDecrement);
				tb_Main.Buffer.Delete(ref iter1, ref iter2);
			}

			if (!ReferenceEquals(m_CaptureStream, null)) {
				try {
					var now = DateTime.UtcNow;
					byte[] toWrite;
					if (now - m_CaptureStamp > TimeSpan.FromMinutes(1)) {
						m_CaptureStamp = now;
						toWrite = Encoding.UTF8.GetBytes(
							String.Format(
								Environment.NewLine + "[Capture {0:0000}-{1:00}-{2:00} {3:00}:{4:00}:{5:00} {6} {7}]" + Environment.NewLine,
								now.Year, now.Month, now.Day,
								now.Hour, now.Minute, now.Second,
								cb_Freq.ActiveText,
								Mode));
						m_CaptureStream.Write(toWrite, 0, toWrite.Length);
					}

					toWrite = Encoding.UTF8.GetBytes(s);
					m_CaptureStream.Write(toWrite, 0, toWrite.Length);
				} catch (Exception ex) {
					Console.WriteLine("Capture failure: {0}: {2}", ex.GetType().Name, ex.Message);
				}
			}
		}
		#endregion

		/// <summary>
		/// Update an entry with a new value, but safely and on the proper thread.
		/// </summary>
		private void SafeUpdateEntryByModem(string entry, string newValue) {
			Application.Invoke(this, new SafeUpdateArgs<string>(entry, newValue), SafeUpdateEntryByModem);
		}

		/// <summary>
		/// The event handler used to do the safe update.
		/// </summary>
		private void SafeUpdateEntryByModem(object sender, EventArgs e) {
			var safe = e as SafeUpdateArgs<string>;
			if (ReferenceEquals(safe, null)) return;

			if (m_ToolbarLayout.Children.Length == 0) return;
			var box = (m_ToolbarLayout.Children.First() as HBox);
			if (ReferenceEquals(box, null)) {
				return;
			}
			var entry = box.Children.Where(c => c is Entry && c.Name.Equals(safe.Target, StringComparison.InvariantCultureIgnoreCase)).FirstOrDefault() as Entry;
			if (ReferenceEquals(entry, null)) {
				return;
			}

			entry.Text = safe.Value;
			entry.ModifyBase(StateType.Normal, m_NormalSettingBase);
			entry.IsEditable = true;

			// next query...
			if (m_OutstandingQueries.Contains(safe.Target)) {
				m_OutstandingQueries.Remove(safe.Target);
				if (m_OutstandingQueries.Count != 0)
					m_Modem.SendCommand(m_OutstandingQueries.First.Value);
			}
		}

		/// <summary>
		/// Attempt to find a settings item that can process the provided command response.
		/// </summary>
		private bool ProcessSettingsBarSettingUpdate(string command, string value) {
			// update the entry
			SafeUpdateEntryByModem(command, value);
			return true;
		}

		/// <summary>
		/// Rebuild the settings bar, based upon the available modem commands.
		/// </summary>
		private void RebuildSettingsBarSafe() {
			Application.Invoke(RebuildSettingsBarHandler);
		}

		/// <summary>
		/// Set the dupe alert state.
		/// </summary>
		private void SetDupe(bool isDupe) {
			tb_Call.ModifyBase(StateType.Normal,
				isDupe
					? new Gdk.Color(255, 255, 0)
					: Gtk.Widget.DefaultStyle.BaseColors[(int)StateType.Normal]);
		}

		/// <summary>
		/// Refresh UI items from configuration.
		/// </summary>
		private void RefreshUiFromConfig() {
			this.Title = "SMI Console - " + m_Config.MyCall;
			if (!ReferenceEquals(m_Config.WindowFont, null)) {
				tb_Main.Font = m_Config.WindowFont.ToFont();
#if LOG_ENTRIES_FOLLOW_WINDOW_STYLE
				foreach (var w in new Widget[] { tbCall, tbRX, tbTX, tbNotes, cbFreq, lCall, lRX, lTX, lNotes, lFreq, bClear.Child as Label, bLog.Child as Label, tbBreakIn.Child as Label, cbFreq.Child as Entry }) {
					w.ModifyFont(m_Config.WindowFont.ToFont());
				}
#endif
			}
			if (!ReferenceEquals(m_Config.WindowForeground, null)) {
				tb_Main.ForeColor = m_Config.WindowForeground.ToColor();
#if LOG_ENTRIES_FOLLOW_WINDOW_STYLE
				foreach (var w in new Widget[] { tbCall, tbRX, tbTX, tbNotes, cbFreq.Child as Entry }) {
					w.ModifyText(StateType.Normal, m_Config.WindowForeground.ToColor());
				}
#endif
			}
			if (!ReferenceEquals(m_Config.WindowForeground, null)) {
				tb_Main.BackColor = m_Config.WindowBackground.ToColor();
#if LOG_ENTRIES_FOLLOW_WINDOW_STYLE
				foreach (var w in new Widget[] { tbCall, tbRX, tbTX, tbNotes, cbFreq.Child as Entry }) {
					w.ModifyBase(StateType.Normal, m_Config.WindowBackground.ToColor());
				}
#endif
			}
		}

		/// <summary>
		/// Test the value of a settings widget, and send it to the modem if it has changed.
		/// </summary>
		/// <param name="w">The widget to consider.</param>
		private void CommitValue(Widget w) {
			Console.WriteLine("SettingsDialog.CommitValue(...)");
			var cmd = w.Name;
			if (String.IsNullOrEmpty(cmd)) return;

			var entry = w as Entry;
			if (!ReferenceEquals(entry, null)) {
				try {
					var newValue = entry.Text.Trim();
					var oldValue = m_SettingsValues[w.Name].Trim();
					if (!String.Equals(newValue, oldValue)) {
						Console.WriteLine("SettingsDialog.CommitValue({0})... was {1}, now {2}", cmd, oldValue, newValue);
						entry.ModifyBase(StateType.Normal, m_Config.WarnColor.ToColor());
						var modem = m_Modem;
						if (ReferenceEquals(modem, null)) return;
						modem.SendCommand(cmd, newValue);
					}
				} catch (Exception ex) {
					Console.WriteLine("SettingsDialog.CommitValue(...) threw {0}: {1}", ex.GetType().Name, ex.Message);
				}
				return;
			}
		}

		/// <summary>
		/// Start the modem, etc.
		/// </summary>
		private void StartupSequence() {
			Console.WriteLine("StartupSequence()");

			// load configuration
			LoadConfig();

			// restart the modem
			RestartModem();

			// configure the update timer
			UpdateTimerTick(this, new EventArgs());
			m_UpdateTimer = new Timer(TimerCallback, null, 1000, 1000);
		}

		/// <summary>
		/// Closes the application gracefully.
		/// </summary>
		private void ShutdownSequence() {
			Console.WriteLine("ShutdownSequence()");

			// save configuration
			SaveConfig();

			if (!ReferenceEquals(m_Modem, null)) {
				m_Modem.Dispose();
				m_Modem = null;
			}
			if (!ReferenceEquals(m_UpdateTimer, null)) {
				m_UpdateTimer.Dispose();
				m_UpdateTimer = null;
			}

#if FAKE_RECEIVER_DATA
			if (!ReferenceEquals(m_TestTimer, null)) {
				m_TestTimer.Dispose();
				m_TestTimer = null;
			}
#endif
		}
		#endregion

		#region Communication Handlers
		/// <summary>
		/// Event handler for incoming modem data stream.
		/// </summary>
		private void ModemDataReceived(object sender, EndpointSMI.DataReceivedEventArgs e) {
#if VERBOSE_DEBUG
			Debug.WriteLine("DEBUG: MainForm.ModemDataReceived({0})", (int)(e.DataByte));
#endif
			// this logic converts [BS][space][BS] into a single back-delete in the TextBox
			if (m_BackspaceState != 0) {
				switch (e.DataByte) {
					case ' ': {
						if (m_BackspaceState == 2) {
							--m_BackspaceState;
							return;
						} else {
							// if we detect that this isn't a three-byte backspace sequence,
							//    stop processing as a backspace sequence, and let the space
							//    fall through to the normal logic
							m_BackspaceState = 0;
							break;
						}
					}
					case BACKSPACE: {
						m_BackspaceState = 0;
						return;
					}
				}
			}

			// everything else flows through here
			switch (e.DataByte) {
				// backspace is a special case
				case BACKSPACE: {
					m_BackspaceState = 2;
					SafeAppendText(BACKSPACE);
				} break;

				// all other bytes; the modem class makes sure the data is printable
				default: {
					SafeAppendText(e.DataByte);
				} break;
			}
		}

		/// <summary>
		/// Event handler for command-value pair received by the modem interface
		/// </summary>
		private void ModemCommandResponse(object sender, EndpointSMI.CommandResponseEventArgs e) {
			Console.WriteLine("DEBUG: MainWindow.ModemCommandResponse({0}, {1}, {2})", e.Command, e.Value, e.Success ? "Success" : "FAIL");

			if (e.Success) {
				switch (e.Command) {
					case @"DOC": {
						RebuildSettingsBarSafe();
					} break;
					default: {
						if (!ProcessSettingsBarSettingUpdate(e.Command, e.Value)) {
							//
							//  TODO: if the settings bar didn't process the value, do something else
							//
						}
					} break;
				}
			}

			//
			//  TODO: there are probably some things we should track here
			//
		}

		/// <summary>
		/// Raised when the link state changes.
		/// </summary>
		private void ModemLinkStateChanged(object sender, LinkStateEventArgs e) {
			Console.WriteLine("DEBUG: MainWindow.ModemLinkStateChanged({0})", e.Connected);

			if (e.Connected) {
				//
				//  TODO: do stuff when the link comes up (but don't send commands yet)
				//
			}
		}

		/// <summary>
		/// Raised when the overall modem state (ready) state changes.
		/// </summary>
		private void ModemOverallStateChanged(object sender, EndpointSMI.ModemStateEventArgs e) {
			Console.WriteLine("DEBUG: MainWindow.ModemOverallStateChanged({0})", e.Ready);
			if (e.Ready) {
				RebuildSettingsBarSafe();
			}
		}
		#endregion

		#region UI Event Handlers
		/// <summary>
		/// Raised by HELP | ABOUT
		/// </summary>
		private void AboutClick(object sender, EventArgs e) {
			MessageBox.Show(this, AboutText, "About " + AppTitle);
		}

		/// <summary>
		/// Raised when the window state changes.
		/// </summary>
		private void WindowStateChangeEvent(object o, WindowStateEventArgs args) {
			this.WindowState = args.Event.NewWindowState;
		}

		/// <summary>
		/// Raised when the QSO window context menu is requested.
		/// </summary>
		private void MainWindowPopulatePopup(object o, PopulatePopupArgs args) {
			Console.WriteLine("MainWindowPopulatePopup(...)");
			int offset = 0;
			MenuItem item = new MenuItem("Call");
			item.Activated += new EventHandler(RightClickCall);
			args.Menu.Insert(item, offset++);
			item = new MenuItem("RX Text");
			item.Activated += new EventHandler(RightClickRX);
			args.Menu.Insert(item, offset++);
			args.Menu.Insert(new SeparatorMenuItem(), offset++);
			item = new MenuItem("Clear");
			item.Activated += new EventHandler(QsoWindowClearClick);
			args.Menu.Insert(item, offset++);
			args.Menu.Insert(new SeparatorMenuItem(), offset++);
			args.Menu.ShowAll();
		}

		/// <summary>
		/// Raised by right-click | CLEAR
		/// </summary>
		private void QsoWindowClearClick(object sender, EventArgs e) {
			tb_Main.Buffer.Clear();
		}

		/// <summary>
		/// Raised when the QSO window context menu's CALL item is clicked.
		/// </summary>
		private void RightClickCall(object sender, EventArgs e) {
			Console.WriteLine("RightClickCall(...)");
			SelectionToCall();
		}

		/// <summary>
		/// Raised when the QSO window context menu's RX item is clicked.
		/// </summary>
		private void RightClickRX(object sender, EventArgs e) {
			Console.WriteLine("RightClickRX(...)");
			SelectionToRX();
		}

		/// <summary>
		/// The delegate method to run on the UI thread to rebuild the settings bar.
		/// </summary>
		private void RebuildSettingsBarHandler(object sender, EventArgs e) {
			// start over
			m_ToolbarLayout.Clear();

			// if no bar commands are configured, just quit
			if (ReferenceEquals(m_Config.ShortcutSettings, null)) return;

			var modem = m_Modem;
			if (ReferenceEquals(modem, null)) return;

			// put a label if no settings available
			if (modem.AvailableCommands == null || modem.AvailableCommands.Length == 0) {
				return;
			}

			// make a list of selected settings that are supported by the current modem
			var toAdd = new LinkedList<string>();
			foreach (var barItem in m_Config.ShortcutSettings) {
				if (m_Modem.AvailableCommands.Contains(barItem)) {
					toAdd.AddLast(barItem);
				}
			}

			// do the layout
			var hbox = new HBox();
			m_ToolbarLayout.PackStart(hbox, false, false, 0);
			m_OutstandingQueries = new LinkedList<string>(toAdd);
			foreach (var barItem in toAdd) {
				Console.WriteLine("Adding Toolbar Item: {0}", barItem);

				var label = new Label(barItem + ":");
				label.SetPadding(8, 0);
				label.SetAlignment(0.9f, 0.5f);
				hbox.Add(label);

				var entry = new Entry();
				m_NormalSettingBase = entry.Style.Base(StateType.Normal);
				entry.Name = barItem;
				entry.WidthChars = 8;
				entry.Activated += SettingActivated;
				entry.FocusInEvent += SettingFocusInEvent;
				entry.FocusOutEvent += SettingFocusOutEvent;
				entry.KeyPressEvent += SpecialNavigationKeyHandler;
				entry.ScrollEvent += SettingsEntryScroll;

				hbox.Add(entry);
				hbox.Add(new VSeparator());
			}
			m_ToolbarLayout.ShowAll();

			if (toAdd.Count != 0) {
				try {
					m_Modem.SendCommand(toAdd.First.Value);
				} catch (Exception ex) {
					Console.WriteLine("RebuildSettingsBarHandler(...): {0}: {1}", ex.GetType().Name, ex.Message);
				}
			}

			Console.WriteLine("RebuildSettingsBarHandler(...): settings bar now has {0} widgets", m_ToolbarLayout.Children.Count());
		}

		/// <summary>
		/// Raised when the mouse scrolls on an Entry.
		/// </summary>
		private void SettingsEntryScroll(object sender, ScrollEventArgs args) {
			var entry = sender as Entry;
			if (ReferenceEquals(entry, null)) return;
			Console.WriteLine("SettingsEntryScroll(...): {0}: {1}: {2}", entry.Name, args.Event.Direction, args.Event);

			var entryText = entry.Text;
			if (String.IsNullOrEmpty(entryText)) return;

			// store the current value, for diff-testing later
			m_SettingsValues[entry.Name] = entry.Text.Trim();

			// compute the increment
			decimal increment = 0;
			switch(args.Event.Direction) {
				case Gdk.ScrollDirection.Up:
					increment = 1m;
					break;
				case Gdk.ScrollDirection.Down:
					increment = -1m;
					break;
				default:
					return;
			}

			// use increment of +/- 0.1 for fractions
			if (entryText.Contains(CultureInfo.CurrentUICulture.NumberFormat.NumberDecimalSeparator)) {
				increment /= 10m;
			}

			// try to read the value, and increment it
			decimal newValue;
			if (Decimal.TryParse(entryText, out newValue)) {
				newValue += increment;
				// neat trick to get decimal places...
				//   ref: https://stackoverflow.com/questions/13477689/find-number-of-decimal-places-in-decimal-value-regardless-of-culture
				int places = BitConverter.GetBytes(Decimal.GetBits(newValue)[3])[2];
				string format = "{0:0";
				if (places > 0)
					format += '.' + new string('0', places);
				format += "}";
				entry.Text = String.Format(format, newValue);
				Console.WriteLine("Change {0} by {1}", entry.Name, increment);
				CommitValue(entry);
			} else {
				// try auto-incrementing in the modem
				var command = entry.Name;
				if (!String.IsNullOrEmpty(command)) {
					var modem = m_Modem;
					if (!ReferenceEquals(modem, null)) {
						string autoArg = args.Event.Direction == Gdk.ScrollDirection.Up ? "+" : "-";
						Console.WriteLine("Modem auto-modify {0} with ", entry.Name, autoArg);
						m_Modem.SendCommand(command, autoArg);
					}
				}
			}
		}

		/// <summary>
		/// Raised when a settings box is exited.
		/// </summary>
		private void SettingFocusOutEvent(object sender, FocusOutEventArgs args) {
			var entry = sender as Entry;
			if (ReferenceEquals(sender, null)) return;
			Console.WriteLine("SettingsDialog.EntryActivated(...): {0}", (sender as Entry).Name);
			CommitValue(entry);
		}

		/// <summary>
		/// Raised when a settings box is focused.
		/// </summary>
		private void SettingFocusInEvent(object sender, FocusInEventArgs args) {
			var entry = sender as Entry;
			if (ReferenceEquals(sender, null)) return;
			Console.WriteLine("SettingsDialog.EntryFocusIn(...): {0}", entry.Name);
			m_SettingsValues[entry.Name] = entry.Text.Trim();
		}

		/// <summary>
		/// Raised when a settings box recieves ENTER.
		/// </summary>
		private void SettingActivated(object sender, EventArgs e) {
			var entry = sender as Entry;
			if (ReferenceEquals(sender, null)) return;
			Console.WriteLine("SettingsDialog.EntryActivated(...): {0}", (sender as Entry).Name);
			CommitValue(entry);
		}

		/// <summary>
		/// Implement auto-scrolling.
		/// </summary>
		private void QsoWindowSizeRequested(object o, SizeRequestedArgs args) {
			var cursor = tb_Main.Buffer.CursorPosition;
			if (cursor == tb_Main.Buffer.CharCount) {
				m_Scroller.Vadjustment.Value = m_Scroller.Vadjustment.Upper - m_Scroller.Vadjustment.PageSize;
			}
		}

		/// <summary>
		/// Raised when text insertion is attempted in the CALL box.
		/// </summary>
		private void CallTextInsertRequest(object sender, FilteredEntryInsertRequestEvent e) {
			// filter out all but alphanumerics, '-' and '/' ... and uppercase the letters
			var sb = new StringBuilder();
			foreach (var ch in e.Text) {
				if (!IsPrint(ch)) continue;
				if (Char.IsWhiteSpace(ch)) continue;
				if (!Char.IsLetterOrDigit(ch) && (ch != '-') && (ch != '/')) continue;
				sb.Append(Char.ToUpper(ch));
			}
			e.Text = sb.ToString();
		}

		/// <summary>
		/// Raised when text insertion is attempted in the RX TEXT box.
		/// </summary>
		private void RxTextInsertRequest(object sender, FilteredEntryInsertRequestEvent e) {
			// filter out all but alphanumerics, '-' and '/' ... and uppercase the letters
			var sb = new StringBuilder();
			var filtered = sender as FilteredEntry;
			foreach (var ch in e.Text) {
				if (!IsPrint(ch)) continue;
				if (ch == ' ') {
					//
					//  TODO: re-write the code to check for duplicate spaces
					//
					sb.Append(' ');
				}
				sb.Append(Char.ToUpper(ch));
			}
			e.Text = sb.ToString();
		}

		/// <summary>
		/// Raised when a macro button is pressed.
		/// </summary>
		private void MacroButtonPressed(object sender, EventArgs e) {
			var button = sender as Button;
			Debug.Assert(!ReferenceEquals(sender, null), "Sender was null");
			var idx = Array.IndexOf(MacroButtons, button);
			if (idx < 0) return;
			Macro macro = m_Config.Macros[idx];
			Debug.Assert(!ReferenceEquals(macro, null));

			// if [ctrl] pressed, edit the macro
			if (m_CtrlDown) {
				m_CtrlDown = false;
				EditMacro(idx);
				return;
			}

			Console.WriteLine(String.Format("DEBUG: Macro Click: Index = {0}", idx));

			// if the macro is empty, give up
			if (String.IsNullOrEmpty(macro.Text)) {
				return;
			}

			// expand the macro
			var textToSend = m_MacroExpander.ExpandRecursive(macro.Text);

			Console.WriteLine("MACRO: Input is : " + macro.Text);
			Console.WriteLine("MACRO: Expansion: " + textToSend);

			// if breakin disabled, give up
			if (BreakInState == BreakIn.Disabled) {
				return;
			}

			// if nothing to send, don't bother
			if (String.IsNullOrEmpty(textToSend)) {
				return;
			}

			// move the QSO cursor to the end
			tb_Main.ScrollToEnd();

			// if the expanded macro doesn't start with [crlf], write a [crlf] first
			switch (textToSend[0]) {
				case '\n':
				case '\r':
					// nop
					break;
				default:
					SafeAppendText('\n');
					break;
			}

			// send the actual macro data
			SendModemData(textToSend);

			// focus the main window, in case we want to type
			if (!tb_Call.HasFocus && !tb_RX.HasFocus) {
				tb_Main.GrabFocus();
			}
		}

		/// <summary>
		/// Raised when the macro expander needs a UI field.
		/// </summary>
		private void MacroFieldRequest(object sender, MacroExpansionField e) {
			switch (e.Key) {
				case MacroKeys.MyCall: e.Value = m_Config.MyCall; break;
				case MacroKeys.Call: e.Value = tb_Call.Text.Trim(); break;
				case MacroKeys.RxExch: e.Value = tb_RX.Text.Trim(); break;
				case MacroKeys.TxExch: e.Value = tb_TX.Text.Trim(); break;
				case MacroKeys.LogCount: e.Value = LogCount().ToString(); break;
				case MacroKeys.LogNext: e.Value = (LogCount() + 1).ToString(); break;
			}
		}

		/// <summary>
		/// Raised when Break-In clicked.
		/// </summary>
		private void BreakInClicked(object sender, EventArgs e) {
			bool bkIn = cb_BreakIn.Active;
			(cb_BreakIn.Child as Label).Text = bkIn ? BreakInOn : BreakInOff;
			Console.WriteLine("BreakInClick: {0}", bkIn);
			tb_Main.GrabFocus();
		}

		/// <summary>
		/// Raised when the CLEAR button is clicked.
		/// </summary>
		private void ClearClicked(object sender, EventArgs e) {
			Console.WriteLine("ClearClick");

			// clear CALL and RX TEXT
			tb_Call.Text = "";
			tb_RX.Text = "";

			// move the QSO cursor to the end
			tb_Main.ScrollToEnd();
			tb_Call.GrabFocus();
		}

		/// <summary>
		/// Raised when the LOG button is clicked.
		/// </summary>
		private void LogClicked(object sender, EventArgs e) {
			Console.WriteLine(
				"LogClick: CALL='{0}', RX='{1}', TX='{2}'",
				tb_Call.Text, tb_RX.Text, tb_TX.Text
			);

			if (String.IsNullOrEmpty(tb_Call.Text)) {
				MessageBox.Show(this, "The CALL field is empty.\n\nEach record must include at least the callsign of the other station.",
					"Log Error");
				return;
			}

			var modem = m_Modem;
			if (ReferenceEquals(modem, null)) {
				MessageBox.Show(
					this,
					"Could not read the mode from the modem.",
					"Log Error",
					MessageType.Warning);
				return;
			}

			// build the new ADIF record
			var adif = new AdifRecord {
				Call = tb_Call.Text.ToUpper().Trim(),
				Freq = (RpcFreq / 1000m).ToString("0.000000"), // MHz, to six decimal places
				Mode = modem.Mode,
				RxString = tb_RX.Text.Trim(),
				TxString = m_MacroExpander.ExpandRecursive(tb_TX.Text.Trim()),
				Notes = tb_Notes.Text.Trim(),
			};
			adif.SetQsoTime(DateTime.UtcNow, m_LastTimeOn);

			// and add it
			bool ok = true;
			try {
				m_Log.AddRecord(adif);
			} catch (Exception) {
				ok = false;
			}

			// if successful
			if (ok) {
				ClearClicked(sender, e);
				m_LogQuery = DateTime.MinValue;
			} else {
				MessageBox.Show(
					this,
					"Could not add log entry: the selected logfile could not be written.",
					"Log Error",
					MessageType.Warning);
			}

			// move the QSO cursor to the end
			tb_Main.ScrollToEnd();

			// go back to the call box
			tb_Call.GrabFocus();
		}

		/// <summary>
		/// Raised when the CALL box text is clicked.
		/// </summary>
		private void CallTextChanged(object sender, EventArgs e) {
			Console.WriteLine("Call now = " + tb_Call.Text);

			// save 'now' as the time-on
			m_LastTimeOn = DateTime.UtcNow;

			// check for duplicates in the log
			CheckDupe();
		}

		/// <summary>
		/// Raised by FILE | SAVE AS...
		/// </summary>
		private void SaveAsClick(object sender, EventArgs e) {
			Console.WriteLine("SaveAsClick(...)");

			using (var dialog = new FileChooserDialog(
					"Save QSO Window",
					this,
					FileChooserAction.Save,
					"Cancel", ResponseType.Cancel,
					"OK", ResponseType.Ok)) {
				// add filters
				var filter = new FileFilter { Name = "Text Files (*.txt)" };
				filter.AddMimeType("text/plain");
				filter.AddPattern("*.txt");
				dialog.AddFilter(filter);
				filter = new FileFilter { Name = "All Files (*.*)" };
				filter.AddPattern("*.*");
				dialog.AddFilter(filter);

				// set initial path
				dialog.SetCurrentFolder(Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments));

				// run the dialog
				var result = (ResponseType)dialog.Run();
				if (result == ResponseType.Ok) {
					try {
						using (var file = File.OpenWrite(dialog.Filename)) {
							var bytes = Encoding.UTF8.GetBytes (tb_Main.Buffer.Text);
							file.Write (bytes, 0, bytes.Length);
						}
					} catch (Exception ex) {
						using (var mb = new MessageDialog(this, DialogFlags.Modal, MessageType.Error, ButtonsType.Ok, "Could not save QSO file: " + ex.Message)) {
							mb.Run ();
							mb.Destroy ();
						}
					}
				}
				dialog.Destroy();
			}
		}

		/// <summary>
		/// Raised by FILE | CAPTURE...
		/// </summary>
		private void CaptureClick(object sender, EventArgs e) {
			Console.WriteLine("CaptureClick(...)");
			using (var dialog = new FileChooserDialog(
					"Capture to File",
					this,
					FileChooserAction.Save,
					"Cancel", ResponseType.Cancel,
					"OK", ResponseType.Ok)) {
				// add filters
				var filter = new FileFilter { Name = "Text Files (*.txt)" };
				filter.AddMimeType("text/plain");
				filter.AddPattern("*.txt");
				dialog.AddFilter(filter);
				filter = new FileFilter { Name = "All Files (*.*)" };
				filter.AddPattern("*.*");
				dialog.AddFilter(filter);

				// set initial path
				dialog.SetCurrentFolder(Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments));

				// run the dialog
				var result = (ResponseType)dialog.Run();
				if (result == ResponseType.Ok) {
					Console.WriteLine ("Capture to " + dialog.Filename);

					if (!ReferenceEquals (m_CaptureStream, null)) {
						try {
							m_CaptureStream.Dispose ();
						} catch { /* nop */
						}
					}

					try {
						m_CaptureStamp = DateTime.UtcNow - TimeSpan.FromHours (1);
						m_CaptureStream = File.Open(dialog.Filename, FileMode.Append, FileAccess.Write);
						m_CaptureStream.Seek(0, SeekOrigin.End);
					} catch (Exception ex) {
						using (var mb = new MessageDialog(this, DialogFlags.Modal, MessageType.Error, ButtonsType.Ok, "Could not open capture file: " + ex.Message)) {
							mb.Run ();
							mb.Destroy ();
						}
					}
				}

				dialog.Destroy();
			}
		}

		/// <summary>
		/// Raised by FILE | CLOSE CAPTURE
		/// </summary>
		private void CloseCaptureClick(object sender, EventArgs e) {
			Console.WriteLine("CloseCaptureClick(...)");
			if (!ReferenceEquals(m_CaptureStream, null)) {
				try { m_CaptureStream.Close(); } catch { }
				try { m_CaptureStream.Dispose(); } catch { }
				m_CaptureStream = null;
			}
		}

		/// <summary>
		/// Raised by FILE | MODEM SETTINGS
		/// </summary>
		private void ModemSettingsClick(object sender, EventArgs e) {
			Console.WriteLine("ModemSettingsClick(...)");

			var modem = m_Modem;
			if (ReferenceEquals(modem, null) || !modem.Ready) {
				MessageBox.Show(this, "The modem is either disconnected or is not ready for settings.", "Modem Not Ready");
				return;
			}

			using (var dialog = new SettingsDialog(this, m_Modem, m_Config)) {
				dialog.Run();
				dialog.Destroy();
				RebuildSettingsBarSafe();
			}
		}

		/// <summary>
		/// Raised by FILE | CONFIGURE SMI
		/// </summary>
		private void SaveConfigNowClick(object sender, EventArgs e) {
			Console.WriteLine("SaveConfigNowClick(...)");

			// just do it
			SaveConfig();
		}

		/// <summary>
		/// Raised by FILE | CONFIGURE SMI
		/// </summary>
		private void ConfigClick(object sender, EventArgs e) {
			Console.WriteLine("ConfigClick(...)");

			using (var dialog = new ConfigDialog(this)) {
				dialog.SetConfiguration(m_Config);
				var result = (ResponseType)dialog.Run();
				int height, width;
				Console.WriteLine("ConfigDialog returned: {0}", result);

				if (result == ResponseType.Ok) {
					m_Config = dialog.GetConfiguration();
					tb_Main.ForeColor = m_Config.WindowForeground.ToColor();
					tb_Main.BackColor = m_Config.WindowBackground.ToColor();
					tb_Main.Font = m_Config.WindowFont.ToFont();
					RefreshUiFromConfig();
					RestartModem();
				}

				dialog.GetSize(out width, out height);
				dialog.Destroy();
			}
		}

		/// <summary>
		/// Raised by LOG | CHOOSE FILE
		/// </summary>
		private void SelectLogFileClick(object sender, EventArgs e) {
			Console.WriteLine("SelectLogFileClick(...)");
			using (var dialog = new FileChooserDialog(
					"Select Log File",
					this,
					FileChooserAction.Save,
					"Cancel", ResponseType.Cancel,
					"OK", ResponseType.Ok)) {
				// add filters
				var filter = new FileFilter { Name = "ADIF Files (*.adi;*.adif)" };
				filter.AddMimeType("text/adif");
				filter.AddPattern("*.adi");
				filter.AddPattern("*.adif");
				dialog.AddFilter(filter);

				// set initial path
				var path = m_Config.LastLogPath;
				if (!File.Exists(path)) {
					path = DefaultLogFile;
				}
				dialog.SetFilename(path);

				// run the dialog
				var result = (ResponseType)dialog.Run();
				if (result == ResponseType.Ok) {
					m_Log = new AdifFile(dialog.Filename);
					m_Config.LastLogPath = dialog.Filename;
				}
				dialog.Destroy();
			}
		}

		/// <summary>
		/// Raised by VIEW | FULL SCREEN
		/// </summary>
		private void FullScreenClick(object sender, EventArgs e) {
			Console.Write("FullScreenClick(...): ");

			// toggle full screen
			if (fs_item.Active) {
				Console.WriteLine("going full screen.");
				this.Fullscreen();
			} else {
				Console.WriteLine("returning from full screen.");
				this.Unfullscreen();
			}
		}

		/// <summary>
		/// Raised by LOG | VIEW
		/// </summary>
		private void LogViewClick(object sender, EventArgs e) {
			Console.WriteLine("LogViewClick(...)");
			if (String.IsNullOrEmpty(m_Config.LoggerPath)) {
				MessageBox.Show(this, "Logger application is not configured.  Check your settings.", "No Editor");
				return;
			}
			if (!File.Exists(m_Config.LoggerPath)) {
				MessageBox.Show(this, "The logger application could not be found.  Check your settings.", "No Editor");
				return;
			}

			// start the logger
			System.Diagnostics.Process.Start(m_Config.LoggerPath, m_Config.LastLogPath);
		}

		/// <summary>
		/// Raised by FILE | EXIT
		/// </summary>
		private void ExitClick(object sender, EventArgs e) {
			Console.WriteLine("ExitClick(...)");
			ShutdownSequence();
			Application.Quit();
		}

		/// <summary>
		/// Raised when the window is shown.
		/// </summary>
		protected override void OnShown() {
			Console.WriteLine("OnShown()");
			StartupSequence();
			base.OnShown();
		}

		/// <summary>
		/// Raised when the window is destroyed.
		/// </summary>
		protected override bool OnDeleteEvent(Gdk.Event evnt) {
			Console.WriteLine("OnDeleteEvent()");
			ShutdownSequence();
			return base.OnDeleteEvent(evnt);
		}

		/// <summary>
		/// Raised when a key is released in the application.
		/// </summary>
		private void MainWindowKeyReleaseEvent(object o, KeyReleaseEventArgs args) {
			switch (args.Event.Key) {
				case Gdk.Key.Control_L:
				case Gdk.Key.Control_R: {
					m_CtrlDown = false;
					//Console.WriteLine("CTRL UP");
				} break;
			}
		}

		/// <summary>
		/// Raised when a key is pressed in the main application.
		/// </summary>
		private void MainWindowKeyPressEvent(object o, KeyPressEventArgs args) {
			switch (args.Event.Key) {
				case Gdk.Key.Control_L:
				case Gdk.Key.Control_R: {
					//Console.WriteLine("CTRL DOWN");
					m_CtrlDown = true;
				} break;
			}
		}

		/// <summary>
		/// This intercepts certain special keys that allow quick navigation.
		/// Certain control keys will cause main input controls to recieve focus.
		/// </summary>
		private void SpecialNavigationKeyHandler(object sender, KeyPressEventArgs e) {
			SpecialNavigationKeyWorker(sender, e.Event);
		}

		/// <summary>
		/// This is the worker for a special/global key event.
		/// </summary>
		private bool SpecialNavigationKeyWorker(object sender, Gdk.EventKey e) {
			// if anybody receives ESC, send it straight to the modem
			if (e.Key == Gdk.Key.Escape && BreakInState == BreakIn.Enabled) {
				m_Modem.SendData(String.Format("{0}", ESC));
				//e.Handled = true;
				return true;
			}

			// if the text boxes receive ENTER, move to the QSO window
			if (ReferenceEquals(sender, tb_Call) || ReferenceEquals(sender, tb_RX) || ReferenceEquals(sender, tb_TX)) {
				if (e.Key == Gdk.Key.Return) {
					tb_Main.GrabFocus();
					//e.Handled = true;
					return true;
				}
			}

			// get modifiers
			bool shift = (e.State & Gdk.ModifierType.ShiftMask) != 0;
			bool ctrl = (e.State & Gdk.ModifierType.ControlMask) != 0;
			bool alt = (e.State & Gdk.ModifierType.Mod1Mask) != 0;

			// handle window full-screen requests
			if (alt && !ctrl && !shift && (e.Key == Gdk.Key.F12)) {
				fs_item.Active = !fs_item.Active;
				return true;
			}

			// handle macro buttons
			if (ctrl) {
				Debug.WriteLine(String.Format(
					"DEBUG: Navigation Handler: {0}{1}{2}{3}",
					ctrl ? "CTRL-" : "",
					alt ? "ALT-" : "",
					shift ? "SHIFT-" : "",
					e.Key.ToString()));
				switch (e.Key) {
					// CTRL-Q -- QSO window
					case Gdk.Key.q:
					case Gdk.Key.Q: {
						tb_Main.GrabFocus();
						return true;
					}

					// CTRL-K -- CALL
					case Gdk.Key.k:
					case Gdk.Key.K: {
						if (alt) {
							SelectionToCall(shift);
						} else {
							tb_Call.GrabFocus();
						}
						return true;
					}

					// CTRL-I -- RX Exchange In
					case Gdk.Key.i:
					case Gdk.Key.I: {
						if (alt) {
							SelectionToRX(shift);
						} else {
							tb_RX.GrabFocus();
						}
						return true;
					}

					// CTRL-T
					case Gdk.Key.t:
					case Gdk.Key.T: {
						tb_TX.GrabFocus();
						return true;
					}

					// CTRL-F
					case Gdk.Key.f:
					case Gdk.Key.F: {
						cb_Freq.GrabFocus();
						return true;
					}

					// CTRL-N
					case Gdk.Key.n:
					case Gdk.Key.N: {
						if (alt) {
							var toAdd = tb_Main.GetSelectedText();
							if (toAdd.Length != 0) {
								if (!shift) {
									tb_Notes.Text = "";
								}
								tb_Notes.Text += toAdd;
								tb_Main.ClearSelection();
							}
						} else {
							tb_Notes.GrabFocus();
						}

						// move the QSO cursor to the end
						tb_Main.ScrollToEnd();
						return true;
					}

					// CTRL-B
					case Gdk.Key.b:
					case Gdk.Key.B: {
						cb_BreakIn.Active = ! cb_BreakIn.Active;
						return true;
					}

					// CTRL-G
					case Gdk.Key.g:
					case Gdk.Key.G: {
						// press the LOG key
						LogClicked(b_Log, new EventArgs());
						return true;
					}

					// CTRL-V
					case Gdk.Key.v:
					case Gdk.Key.V: {
						// since the main text box is r/o, simulate paste
						if (ReferenceEquals(tb_Main, sender)) {
							RightClickPaste(tb_Main, new EventArgs());
						}

						// move the QSO cursor to the end
						tb_Main.ScrollToEnd();
						return true;
					}

					// CTRL-W
					case Gdk.Key.w:
					case Gdk.Key.W: {
						// get the current call contents
						var call = tb_Call.Text.Trim();
						Console.WriteLine("DEBUG: \"{0}\" Who?", call);

						// don't try if the bk-in is disabled
						if (BreakInState != BreakIn.Enabled) return true;

						// form macro text
						var macro = call + "?";

						// send data to the modem
						SafeAppendText("\n");
						SendModemData(macro);
						return true;
					}

					// CTRL-# - select setting
					case Gdk.Key.Key_1:
					case Gdk.Key.Key_2:
					case Gdk.Key.Key_3:
					case Gdk.Key.Key_4:
					case Gdk.Key.Key_5:
					case Gdk.Key.Key_6:
					case Gdk.Key.Key_7:
					case Gdk.Key.Key_8:
					case Gdk.Key.Key_9: {
						int idx = e.Key - Gdk.Key.Key_1;
						int i = 0;
						foreach (var widget in (m_ToolbarLayout.Children.FirstOrDefault() as Container).Children.Where(c => !(c is Label) && !(c is VSeparator))) {
							if (i++ == idx) {
								widget.GrabFocus();
								break;
							}
						}
					} break;

					// CTRL-F* - edit macro
					case Gdk.Key.F1:
					case Gdk.Key.F2:
					case Gdk.Key.F3:
					case Gdk.Key.F4:
					case Gdk.Key.F5:
					case Gdk.Key.F6:
					case Gdk.Key.F7:
					case Gdk.Key.F8:
					case Gdk.Key.F9:
					case Gdk.Key.F10:
					case Gdk.Key.F11:
					case Gdk.Key.F12: {
						if (!alt && !shift) {
							Console.WriteLine("Macro Keyboard EDIT: " + e.Key.ToString());
							var idx = e.Key - Gdk.Key.F1;
							EditMacro(idx);
						}
						return true;
					}
				}
			}

			if (!ctrl && !alt && !shift) {
				switch (e.Key) {
					case Gdk.Key.F1:
					case Gdk.Key.F2:
					case Gdk.Key.F3:
					case Gdk.Key.F4:
					case Gdk.Key.F5:
					case Gdk.Key.F6:
					case Gdk.Key.F7:
					case Gdk.Key.F8:
					case Gdk.Key.F9:
					case Gdk.Key.F10:
					case Gdk.Key.F11:
					case Gdk.Key.F12: {
						Console.WriteLine("Macro Keyboard RUN: " + e.Key.ToString());
						var idx = e.Key - Gdk.Key.F1;
						MacroButtonPressed(MacroButtons[idx], new EventArgs());
						return true;
					}
				}
			}
			return false;
		}

		/// <summary>
		/// Move the selected text to the RX field.
		/// </summary>
		private void SelectionToRX(bool shift = false) {
			var toAdd = tb_Main.GetSelectedText();
			if (toAdd.Length != 0) {
				if (!shift) {
					tb_RX.Text = "";
				}
				tb_RX.Text += FilterExchange(toAdd);
				tb_Main.ClearSelection();

				// move the QSO cursor to the end
				tb_Main.ScrollToEnd();
			}
		}

		/// <summary>
		/// Move the selected text to the CALL field.
		/// </summary>
		private void SelectionToCall(bool shift = false) {
			var toAdd = tb_Main.GetSelectedText();
			if (toAdd.Length != 0) {
				if (!shift) {
					tb_Call.Text = "";
				}
				tb_Call.Text += FilterCall(toAdd);
				tb_Main.ClearSelection();

				// move the QSO cursor to the end
				tb_Main.ScrollToEnd();
			}
		}

		/// <summary>
		/// Edit a specific macro.
		/// </summary>
		/// <param name="idx">The macro index, zero-based.</param>
		private void EditMacro(int idx) {
			// pop-up a modal macro editor and run it
			using (var editor = new MacroEditor(this)) {
				// fill in the macro data
				editor.MacroLabel = m_Config.Macros[idx].Label;
				editor.MacroText = m_Config.Macros[idx].Text;
				// configure the editor colors
				editor.Foreground = m_Config.WindowForeground.ToColor();
				editor.Background = m_Config.WindowBackground.ToColor();
				// run the editor
				var result = (ResponseType)(editor.Run());
				// if the user clicked OK...
				if (result == ResponseType.Ok) {
					// save the macro changes
					m_Config.Macros[idx].Label = editor.MacroLabel;
					m_Config.Macros[idx].Text = editor.MacroText;
					// and update the button label
					MacroButtons[idx].Label = editor.MacroLabel;
				}
				// clean up
				editor.Destroy();
			}
		}

		/// <summary>
		/// UI update timer.
		/// </summary>
		private void UpdateTimerTick(object sender, EventArgs e) {
			//Console.WriteLine("UpdateTimerTick()");

			var now = DateTime.UtcNow;

			// if it has been a while since the last frequency query, run one
			if ((now - LastFrequencyQuery) >= FreqInterval) {
				StartRpcFreqQuery(); // read XMLRPC frequency
			}

			// update the log count
			if ((now - m_LogQuery) >= LogInterval) {
				m_LogCount = m_Log.Count;
			}

			//
			//  TODO: FIXME: do all of these tasks in response to explicit modem events, rather than polled on a timer
			//

			//
			//  CONNECTION STATE
			//
			var modem = m_Modem;
			var readyText = "Disconnected";
			bool ready = false;
			if (!ReferenceEquals(modem, null)) {
				try {
					ready = modem.Ready;
					if (ready)
						readyText = "Ready";
				} catch {
					// nop
				}
			}

			//
			//  MODE
			//
			var modeText = "";
			if (!ReferenceEquals(modem, null)) {
				try {
					modeText = modem.Mode;
				} catch {
					// nop
				}
			}

			//
			//  VERSION
			//
			var versionText = "";
			if (!ReferenceEquals(modem, null)) {
				try {
					versionText = modem.Version;
				} catch {
					// nop
				}
			}

			//
			//  CAPTURE
			//
			var capText = "";
			var cap = m_CaptureStream;
			if (!ReferenceEquals(cap, null)) {
				try {
					capText = "Capture: " + System.IO.Path.GetFileName(cap.Name);
				} catch {
					// nop
				}
			}

			var sb = new StringBuilder();
			sb.Append(readyText);
			if (!String.IsNullOrEmpty(modeText)) {
				sb.Append(" - ");
				sb.Append(modeText);
			}
			if (!String.IsNullOrEmpty(versionText)) {
				sb.Append(" - ");
				sb.Append(versionText);
			}
			if (!String.IsNullOrEmpty(capText)) {
				sb.Append(" - ");
				sb.Append(capText);
			}
			if (m_LogCount >= 0) {
				sb.Append(" - ");
				sb.Append(String.Format("Log Count: {0}", m_LogCount));
			}
			sb_Status.Push(0, sb.ToString());
		}

		/// <summary>
		/// Handle keyboard key-press in main window.
		/// </summary>
		private void QsoWindowKeyPress(object sender, UnfilteredKeyPressEventArgs args) {
			Console.WriteLine("QsoWindowKeyPress({0})", args.Event.KeyValue);
			if (args.Event.Key == Gdk.Key.Return) {
				SendData(Convert.ToChar(13));
				SendData(Convert.ToChar(10));
				args.CancelEvent = true;
				return;
			}

			if (args.Event.Key == Gdk.Key.Tab) {
				tb_Call.GrabFocus();
				args.CancelEvent = true;
				return;
			}

			if (SpecialNavigationKeyWorker(sender, args.Event)) {
				args.CancelEvent = true;
				return;
			}

			bool modifiersOk = (args.Event.State & (Gdk.ModifierType.ControlMask | Gdk.ModifierType.Mod1Mask)) == 0;
			//Console.WriteLine("Modifiers: {0}; modOk; {1}", args.Event.State, modifiersOk);
			if (args.Event.KeyValue < 128 && modifiersOk) {
				char ch = Convert.ToChar(args.Event.KeyValue);	// convert the key to an ascii value
				args.CancelEvent = true;						// consume this event
				tb_Main.ScrollToEnd();
				m_Scroller.Vadjustment.Value = m_Scroller.Vadjustment.Upper - m_Scroller.Vadjustment.PageSize;

				if (IsPrint(ch)) {
					SendData(ch);
				}
			}
		}

		/// <summary>
		/// Raised when the frequency/band is changed.
		/// </summary>
		private void cbFreq_Changed(object sender, EventArgs e) {
			CheckDupe();
		}

		/// <summary>
		/// Right-click | PASTE handler.
		/// </summary>
		private void RightClickPaste(object sender, EventArgs e) {
			Console.WriteLine("RightClickPaste()");

			//
			//  TODO:
			//
		}
		#endregion

		#region State Persistence
		/// <summary>
		/// Restore state from config file.
		/// </summary>
		private void LoadConfig() {
			try {
				using (var f = File.OpenText(ConfigFile)) {
					m_Config = s_ConfigSerializer.Deserialize(f) as Config;
				}
			} catch (DirectoryNotFoundException) {
				MessageBox.Show(this, "The SMI configuration was not found, so a new configuration will be created.", "SMI Configuration");
			} catch (FileNotFoundException) {
				MessageBox.Show(this, "The SMI configuration was not found, so a new configuration will be created.", "SMI Configuration");
			} catch (Exception ex) {
				MessageBox.Show(this, String.Format("While reading {0}: {1}: {2}: {3}",
					ConfigFile, ex.GetType().Name, ex.Message, ex.StackTrace), "Error");
			}

			// restore macros
			if (ReferenceEquals(m_Config.Macros, null)) {
				m_Config.Macros = new Macro[MacroButtons.Length];
				for (int i = 0; i != MacroButtons.Length; ++i) {
					string label = "F" + (i + 1);
					m_Config.Macros[i] = new Macro() { Label = label };
				}
			}
			int buttons = Math.Min(m_Config.Macros.Length, MacroButtons.Length);
			for (int i = 0; i != buttons; ++i) {
				MacroButtons[i].Label = m_Config.Macros[i].Label;
				if (i >= m_Config.MacroButtonCount) {
					MacroButtons[i].Visible = false;
				}
			}

			// restore TX exchange
			tb_TX.Text = m_Config.TxText ?? "";

			// configure logging
			m_Log = new AdifFile(String.IsNullOrEmpty(m_Config.LastLogPath) ? DefaultLogFile : m_Config.LastLogPath);

			// restore UI items
			if (m_Config.WindowFont != null) {
				tb_Main.Font = m_Config.WindowFont.ToFont();
			}
			if (m_Config.WindowForeground != null) {
				tb_Main.ForeColor = m_Config.WindowForeground.ToColor();
			}
			if (m_Config.WindowBackground != null) {
				tb_Main.BackColor = m_Config.WindowBackground.ToColor();
			}

			// refresh UI elements
			RefreshUiFromConfig();
		}

		/// <summary>
		/// Save state to config file.
		/// </summary>
		private void SaveConfig() {
			m_Config.TxText = tb_TX.Text.Trim();
			m_Config.LastLogPath = m_Log.Path;
			m_Config.WindowFont = KK5JY.Config.ConfigUtils.CreateSerializableFont(tb_Main.Font);
			m_Config.WindowForeground = ConfigUtils.CreateSerializableColor(tb_Main.ForeColor);
			m_Config.WindowBackground = ConfigUtils.CreateSerializableColor(tb_Main.BackColor);

			try {
				if (!Directory.Exists(ConfigPath)) {
					Directory.CreateDirectory(ConfigPath);
				}
				using (var f = File.Create(ConfigFile)) {
					s_ConfigSerializer.Serialize(f, m_Config);
				}
			} catch {
				// nop
			}
		}
		#endregion

		#region XMLRPC Stuff
		private decimal RpcFreq {
			get {
				string working = cb_Freq.ActiveText.ToLower().Trim();
				if (working.EndsWith("m")) {
					working = working.Substring(0, working.Length - 1);
					switch (working) {
						case "160": return 1800.0m;
						case "80": return 3500.0m;
						case "40": return 7000.0m;
						case "30": return 10000.0m;
						case "20": return 14000.0m;
						case "17": return 18100.0m;
						case "15": return 21000.0m;
						case "12": return 24890.0m;
						case "10": return 28000.0m;
						case "6": return 50000.0m;
						case "2": return 144000.0m;
					}
				}
				decimal freq;
				if (Decimal.TryParse(working, out freq)) {
					return freq;
				}
				return 0;
			}

			set {
				cb_Freq.RemoveText(0);
				cb_Freq.InsertText(0, value.ToString("0.000"));
				cb_Freq.Active = 0;
			}
		}

		private void SetRpcFreq(object sender, EventArgs args) {
			var fargs = args as SetFrequencyArgs;
			Console.WriteLine("SetRpcFreq({0:0.000})", fargs.Frequency);
			RpcFreq = fargs.Frequency;
		}

		private void RpcGetFreqCallback(IAsyncResult ar) {
			GetDecimalDelegate d = ar.AsyncState as GetDecimalDelegate;
			Debug.Assert(!ReferenceEquals(d, null));
			if (d == null) return;
			var f = d.EndInvoke(ar);
			if (f > 0m) {
				Application.Invoke(this, new SetFrequencyArgs { Frequency = (f / 1000) }, SetRpcFreq);
			}
		}

		private void StartRpcFreqQuery() {
			GetDecimalDelegate d = m_Rig.GetFrequency;
			d.BeginInvoke(RpcGetFreqCallback, d);
			LastFrequencyQuery = DateTime.UtcNow;
		}
		#endregion
	}
}
