﻿/*
 *
 *
 *    SettingsDialog.cs
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2017 by Matthew K. Roberts, KK5JY. All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *    
 *
 */

using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Text;
using System.IO;

using Gtk;

using KK5JY.Config;
using KK5JY.IO;

namespace KK5JY.SMI.GTK {
	/// <summary>
	/// A dialog for changing modem settings (WPM, PTT timing, etc).
	/// </summary>
	public sealed class SettingsDialog : Dialog {
		#region Private Fields
		private EndpointSMI m_Modem;
		private Config m_Config;
		private Entry[] m_SettingsEntries;
		private string[] m_SettingsCommands;
		private string[] m_SettingsValues;
		private int m_SettingsIndex;
		private Gdk.Color m_NormalBase;
		private readonly string[] ReadOnlyCommands = new[] { @"ECHO", @"MODE", @"VERSION" };
		private Table m_Table;
		#endregion

		#region Public Properties
		/// <summary>
		/// Get or set the list of commands that are selected.
		/// </summary>
		/// <remarks>
		/// This proeprty uses the CheckButton widgets for storage, and sets/gets the array of strings on the fly.
		/// </remarks>
		public IEnumerable<string> SelectedCommands {
			get {
				foreach (var check in m_Table.Children.Where(c => c is CheckButton).Cast<CheckButton>()) {
					if (check.Active) {
						yield return check.Name;
					}
				}
			}
			set {
				if (ReferenceEquals(m_Table, null) || ReferenceEquals(m_Table.Children, null)) return;
				var checks = m_Table.Children.Where(c => c is CheckButton).Cast<CheckButton>().ToList();
				foreach (var check in checks) {
					if (value.Where(v => v.Equals(check.Name, StringComparison.InvariantCultureIgnoreCase)).Count() != 0) {
						check.Active = true;
					} else {
						check.Active = false;
					}
				}
			}
		}
		#endregion

		#region Constructors
		/// <summary>
		/// Construct a new <c>SettingsDialog</c>.
		/// </summary>
		/// <param name="parent">The parent window.</param>
		/// <param name="modem">The active modem instance.</param>
		/// <param name="config">The active configuration.</param>
		public SettingsDialog(Window parent, EndpointSMI modem, Config config)
			: base("Modem Settings", parent, Gtk.DialogFlags.DestroyWithParent) {
			// save the modem reference
			m_Modem = modem;
			m_Config = config;

			// configure the window itself
			this.Title = "Modem Settings";
			this.Deletable = true;
			this.Resizable = false;
			this.Modal = true;

			var ready = modem.Ready;
			if (ready && (modem.AvailableCommands != null)) {
				var commands = new LinkedList<string>(modem.AvailableCommands.OrderBy(s => s));
				foreach (var special in ReadOnlyCommands) {
					if (commands.Contains(special)) {
						commands.Remove(special);
						commands.AddFirst(special);
					}
				}
				m_SettingsEntries = new Entry[commands.Count];
				m_SettingsValues = new string[commands.Count];
				m_SettingsCommands = commands.ToArray();

				const int columns = 3;
				m_Table = new Table(Convert.ToUInt32(m_SettingsCommands.Length), columns, false);
				this.VBox.PackStart(m_Table, false, true, 0);
				uint row = 0;

				Widget selected = null;
				foreach (var command in commands) { // for each row in the window
					var label = new Label(command); // the label for the row
					label.SetAlignment(0.9f, 0.5f);
					label.SetPadding(5, 0);
					var entry = new Entry(); // the entry for the row
					entry.IsEditable = false; // assume all Entry's are r/o initially
					CheckButton check = null; // assume the row doesn't get a checkbox, either
					switch (command.ToUpper()) {
						case "MODE":
						case "VERSION": {
							entry.ModifyBase(StateType.Normal, label.Style.Background(StateType.Normal));
							entry.Text = command.ToUpper() == "VERSION" ? modem.Version : modem.Mode;
							entry.Name = command.ToUpper();
						} break;
						case "ECHO": {
							entry.ModifyBase(StateType.Normal, label.Style.Background(StateType.Normal));
							entry.Name = command.ToUpper();
						} break;
						default: {
							entry.ModifyBase(StateType.Normal, label.Style.Background(StateType.Normal));
							if (ReferenceEquals(selected, null)) {
								selected = entry;
								m_NormalBase = entry.Style.Base(StateType.Normal);
							}
							entry.Name = command.ToUpper();
							entry.Activated += EntryActivated;
							entry.FocusInEvent += EntryFocusIn;
							entry.FocusOutEvent += EntryFocusOut;
							entry.ScrollEvent += SettingsEntryScroll;
							check = new CheckButton();
							check.Name = command.ToUpper();
							check.Toggled += CheckChanged;
						} break;
					}
					m_Table.Attach(label, 0, 1, row, row + 1);
					m_Table.Attach(entry, 1, 2, row, row + 1);
					if (!ReferenceEquals(check, null))
						m_Table.Attach(check, 2, 3, row, row + 1);
					m_SettingsEntries[row] = entry;
					++row;
				}
				if (!ReferenceEquals(selected, null)) {
					selected.GrabFocus();
				}
			} else {
				this.VBox.PackStart(new Label("The modem is not ready."));
			}

			if (!ReferenceEquals(m_Config.ShortcutSettings, null)) {
				this.SelectedCommands = m_Config.ShortcutSettings;
			}

			// fill the action area buttons
			var import = new Button();
			import.Label = "Import";
			import.Pressed += this.ImportModemSettingsClick;
			ActionArea.PackStart(import);
			var export = new Button();
			export.Label = "Export";
			export.Pressed += this.ExportModemSettingsClick;
			ActionArea.PackStart(export);

			this.AddButton("Done", ResponseType.Ok);
			this.SetResponseSensitive(ResponseType.Ok, true);
			this.DefaultResponse = ResponseType.Ok;

			ShowAll();
		}
		#endregion

		#region Utilities
		/// <summary>
		/// Update an entry with a new value, but safely and on the proper thread.
		/// </summary>
		private void SafeUpdateEntryByModem(Entry entry, string newValue) {
			Application.Invoke(this, new SafeUpdateArgs<Entry>(entry, newValue), SafeUpdateEntryByModem);
		}

		/// <summary>
		/// The event handler used to do the safe update.
		/// </summary>
		private void SafeUpdateEntryByModem(object sender, EventArgs e) {
			var safe = e as SafeUpdateArgs<Entry>;
			if (ReferenceEquals(safe, null)) return;
			safe.Target.Text = safe.Value;
			if (!ReadOnlyCommands.Contains(safe.Target.Name.ToUpper())) {
				safe.Target.ModifyBase(StateType.Normal, m_NormalBase);
				safe.Target.IsEditable = true;
			}
		}

		/// <summary>
		/// Test the value of a settings widget, and send it to the modem if it has changed.
		/// </summary>
		/// <param name="w">The widget to consider.</param>
		private void CommitValue(Widget w) {
			Console.WriteLine("SettingsDialog.CommitValue(...)");
			var cmd = w.Name;
			if (String.IsNullOrEmpty(cmd)) return;
			var idx = Array.IndexOf(m_SettingsCommands, cmd);
			if (idx < 0) return;

			var entry = w as Entry;
			if (!ReferenceEquals(entry, null)) {
				try {
					var newValue = entry.Text.Trim();
					var oldValue = m_SettingsValues[idx].Trim();
					if (!String.Equals(newValue, oldValue)) {
						Console.WriteLine("SettingsDialog.CommitValue({0})... was {1}, now {2}", cmd, oldValue, newValue);
						entry.ModifyBase(StateType.Normal, m_Config.WarnColor.ToColor());
						m_Modem.SendCommand(cmd, newValue);
					}
				} catch (Exception ex) {
					Console.WriteLine("SettingsDialog.CommitValue(...) threw {0}: {1}", ex.GetType().Name, ex.Message);
				}
				return;
			}
		}
		#endregion

		#region Event Handlers
		/// <summary>
		/// Raised when the window is shown.
		/// </summary>
		protected override void OnShown() {
			m_Modem.CommandResponse += ModemCommandResponse;
			if (ReferenceEquals(m_SettingsCommands, null)) return;
			m_SettingsIndex = 0;
			var cmd = m_SettingsCommands[m_SettingsIndex];
			m_Modem.SendCommand(cmd);
			Console.WriteLine("SettingsDialog Query: {0}", cmd);
			base.OnShown();
		}

		/// <summary>
		/// Raised when the window is closed.
		/// </summary>
		protected override void OnDestroyed() {
			m_Modem.CommandResponse -= ModemCommandResponse;
			base.OnDestroyed();
		}

		/// <summary>
		/// Raised when a checkbox is changed.
		/// </summary>
		private void CheckChanged(object sender, EventArgs e) {
			var check = sender as CheckButton;
			if (ReferenceEquals(check, null)) return;
			m_Config.ShortcutSettings = SelectedCommands.ToArray();
			var settingsString = String.Join(" ", m_Config.ShortcutSettings);
			Console.WriteLine("Check[{0}] = {1}; Selected settings[{2}] now: {3}", check.Name, check.Active, SelectedCommands.Count(), settingsString);
		}

		/// <summary>
		/// Raised by FILE | IMPORT.
		/// </summary>
		private void ImportModemSettingsClick(object sender, EventArgs e) {
			Console.WriteLine("ImportModemSettingsClick(...)");

			// sanity check
			var modem = m_Modem;
			if (ReferenceEquals(modem, null)) {
				MessageBox.Show(this, "The modem settings could not be read; the modem is unavailable.", "Export Error");
				return;
			}

			// ask for a file name
			string file = null;
			using (var dialog = new FileChooserDialog(
					"Load Modem Settings from File",
					this,
					FileChooserAction.Save,
					"Cancel", ResponseType.Cancel,
					"OK", ResponseType.Ok)) {
				// add filters
				var filter = new FileFilter { Name = "Text Files (*.txt)" };
				filter.AddMimeType("text/plain");
				filter.AddPattern("*.txt");
				dialog.AddFilter(filter);
				filter = new FileFilter { Name = "All Files (*.*)" };
				filter.AddPattern("*.*");
				dialog.AddFilter(filter);

				// set initial path
				dialog.SetCurrentFolder(Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments));

				// run the dialog
				var result = (ResponseType)dialog.Run();
				if (result == ResponseType.Ok) {
					file = dialog.Filename;
				}
				dialog.Destroy();
			}

			if (String.IsNullOrEmpty(file))
				return;

			if (!File.Exists(file)) {
				MessageBox.Show(this, "The settings file could not be found", "Import Error");
				return;
			}

			// import from that file
			try {
				using (var fs = File.OpenText(file)) {
					while (!fs.EndOfStream) {
						var line = fs.ReadLine().Trim();
						if (String.IsNullOrEmpty(line)) continue;
						if (line[0] == '#') continue;
						var parts = line.Split(new[] { '=' }, 2, StringSplitOptions.RemoveEmptyEntries);
						if (parts.Length != 2) continue;
						var cmd = parts[0].ToUpper();
						var value = parts[1];
						if (ReadOnlyCommands.Contains(cmd))
							continue;
						var idx = Array.IndexOf(m_SettingsCommands, cmd);
						if (idx < 0) continue;
						//
						//  TODO: this is going to dump a lot of data at the modem; this needs
						//        to be queued, so that only one outstanding command is pending
						//        at a time...
						//
						m_SettingsEntries[idx].Text = value;
						CommitValue(m_SettingsEntries[idx]);
					} 
				}
			} catch (Exception ex) {
				MessageBox.Show(this, String.Format("Could not read settings from file: {0}: {1}", ex.GetType().Name, ex.Message), "Export Error");
			}
		}

		/// <summary>
		/// Raised by FILE | EXPORT.
		/// </summary>
		private void ExportModemSettingsClick(object sender, EventArgs e) {
			Console.WriteLine("ExportModemSettingsClick(...)");

			// sanity check
			var modem = m_Modem;
			if (ReferenceEquals(modem, null)) {
				MessageBox.Show(this, "The modem settings could not be read; the modem is unavailable.", "Export Error");
				return;
			}

			// ask for a file name
			string file = null;
			using (var dialog = new FileChooserDialog(
					"Save Modem Settings to File",
					this,
					FileChooserAction.Save,
					"Cancel", ResponseType.Cancel,
					"OK", ResponseType.Ok)) {
				// add filters
				var filter = new FileFilter { Name = "Text Files (*.txt)" };
				filter.AddMimeType("text/plain");
				filter.AddPattern("*.txt");
				dialog.AddFilter(filter);
				filter = new FileFilter { Name = "All Files (*.*)" };
				filter.AddPattern("*.*");
				dialog.AddFilter(filter);

				// set initial path
				dialog.SetCurrentFolder(Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments));

				// run the dialog
				var result = (ResponseType)dialog.Run();
				if (result == ResponseType.Ok) {
					file = dialog.Filename;
				}
				dialog.Destroy();
			}

			// export to that file
			if (!String.IsNullOrEmpty(file)) {
				if (File.Exists(file)) {
					try {
						if (MessageBox.Show(this, "The file already exists.  Overwrite?", "Confirm", MessageType.Warning, ButtonsType.YesNo) != ResponseType.Yes) {
							return;
						}
						File.Delete(file);
					} catch (Exception ex) {
						Console.WriteLine("WARNING: could not remove file {0}: {1}: {2}", file, ex.GetType().Name, ex.Message);
					}
				}
				try {
					using (var fs = File.OpenWrite(file)) {
						using (var sw = new StreamWriter(fs)) {
							for (int i = 0; i != m_SettingsCommands.Length; ++i) {
								if (!String.IsNullOrEmpty(m_SettingsValues[i])) {
									if (ReadOnlyCommands.Contains(m_SettingsCommands[i])) {
										sw.Write("# ");
									}
									sw.WriteLine("{0}={1}", m_SettingsCommands[i], m_SettingsValues[i]);
								}
							}
						}
					}
				} catch (Exception ex) {
					MessageBox.Show(this, String.Format("Could not write settings to file: {0}: {1}", ex.GetType().Name, ex.Message), "Export Error");
				}
			}
		}

		/// <summary>
		/// Raised when the mouse scrolls on an Entry.
		/// </summary>
		private void SettingsEntryScroll(object sender, ScrollEventArgs args) {
			var entry = sender as Entry;
			if (ReferenceEquals(entry, null)) return;
			Console.WriteLine("SettingsEntryScroll(...): {0}: {1}: {2}", entry.Name, args.Event.Direction, args.Event);

			var entryText = entry.Text;
			if (String.IsNullOrEmpty(entryText)) return;

			// store the current value, for diff-testing later
			var idx = Array.IndexOf(m_SettingsCommands, entry.Name);
			m_SettingsValues[idx] = entry.Text.Trim();

			// compute the increment
			decimal increment = 0;
			switch (args.Event.Direction) {
				case Gdk.ScrollDirection.Up:
				increment = 1m;
				break;
				case Gdk.ScrollDirection.Down:
				increment = -1m;
				break;
				default:
				return;
			}

			// use increment of +/- 0.1 for fractions
			if (entryText.Contains(CultureInfo.CurrentUICulture.NumberFormat.NumberDecimalSeparator)) {
				increment /= 10m;
			}

			// try to read the value, and increment it
			decimal newValue;
			if (Decimal.TryParse(entryText, out newValue)) {
				newValue += increment;
				// neat trick to get decimal places...
				//   ref: https://stackoverflow.com/questions/13477689/find-number-of-decimal-places-in-decimal-value-regardless-of-culture
				int places = BitConverter.GetBytes(Decimal.GetBits(newValue)[3])[2];
				string format = "{0:0";
				if (places > 0)
					format += '.' + new string('0', places);
				format += "}";
				entry.Text = String.Format(format, newValue);
				entry.GrabFocus();
				Console.WriteLine("Change {0} by {1}", entry.Name, increment);
				CommitValue(entry);
			} else {
				// try auto-incrementing in the modem
				var command = entry.Name;
				if (!String.IsNullOrEmpty(command)) {
					var modem = m_Modem;
					if (!ReferenceEquals(modem, null)) {
						string autoArg = args.Event.Direction == Gdk.ScrollDirection.Up ? "+" : "-";
						Console.WriteLine("Modem auto-modify {0} with ", entry.Name, autoArg);
						m_Modem.SendCommand(command, autoArg);
					}
				}
			}
		}

		/// <summary>
		/// Raised when the modem sends use a command/response message.
		/// </summary>
		private void ModemCommandResponse(object sender, EndpointSMI.CommandResponseEventArgs e) {
			try {
				if (!e.Success) return;

				// update the corresponding settings box
				var idx = Array.IndexOf(m_SettingsCommands, e.Command);
				if (idx >= 0) {
					var newValue = e.Value;
					var entry = m_SettingsEntries[idx];
					SafeUpdateEntryByModem(entry, newValue);
					m_SettingsValues[idx] = e.Value ?? "";
				}

				// prime the next settings box, if there are any left
				if (m_SettingsIndex < m_SettingsCommands.Length - 1) {
					++m_SettingsIndex;
					var cmd = m_SettingsCommands[m_SettingsIndex];
					m_Modem.SendCommand(cmd);
					Console.WriteLine("SettingsDialog Query: {0}", cmd);
				}
			} catch (Exception ex) {
				Console.WriteLine(
					"SettingsDialog.ModemCommandResponse({0}:{1}): {2}: {3}",
					e.Command,
					e.Value ?? "",
					ex.GetType().Name,
					ex.Message
				);
			}
		}

		/// <summary>
		/// Raised when a settings Entry widget receives the ENTER key.
		/// </summary>
		private void EntryActivated(object sender, EventArgs e) {
			var entry = sender as Entry;
			if (ReferenceEquals(sender, null)) return;
			Console.WriteLine("SettingsDialog.EntryActivated(...): {0}", (sender as Entry).Name);
			CommitValue(entry);
		}

		/// <summary>
		/// Raised when a settings Entry widget is entered.
		/// </summary>
		private void EntryFocusIn(object sender, EventArgs args) {
			var entry = sender as Entry;
			if (ReferenceEquals(sender, null)) return;
			Console.WriteLine("SettingsDialog.EntryFocusIn(...): {0}", entry.Name);
			var idx = Array.IndexOf(m_SettingsCommands, entry.Name);
			if (idx >= 0) {
				m_SettingsValues[idx] = entry.Text.Trim();
			}
		}

		/// <summary>
		/// Raised when a settings Entry widget is exited.
		/// </summary>
		private void EntryFocusOut(object sender, EventArgs args) {
			var entry = sender as Entry;
			if (ReferenceEquals(sender, null)) return;
			Console.WriteLine("SettingsDialog.EntryFocusOut(...): {0}", (sender as Entry).Name);
			CommitValue(entry);
		}
		#endregion
	}
}
