﻿/*
 *
 *
 *    TerminalWindow.cs
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2017 by Matthew K. Roberts, KK5JY. All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *    
 *
 */

using System;
using Gtk;

namespace KK5JY.SMI.GTK {
	/// <summary>
	/// Raised when a key is pressed in a <c>TerminalWindow</c>.
	/// </summary>
	public sealed class UnfilteredKeyPressEventArgs : EventArgs {
		/// <summary>
		/// The key that was pressed.
		/// </summary>
		public Gdk.EventKey Event { get; internal set; }

		/// <summary>
		/// Flag indicating that the framework should consume the vent,
		/// and not pass it to the underlying window type.
		/// </summary>
		public bool CancelEvent { get; set; }

		/// <summary>
		/// Constructor.
		/// </summary>
		public UnfilteredKeyPressEventArgs() {
			CancelEvent = false;
		}
	}

	/// <summary>
	/// A terminal window.
	/// </summary>
	public class TerminalWindow : TextView {
		#region Public Properties
		/// <summary>
		/// Get the word nearest the most recent right-click (popup) event.
		/// </summary>
		public string PopupNearestWord { get; private set; }

		/// <summary>
		/// The window font.
		/// </summary>
		public Pango.FontDescription Font {
			get { return this.Style.FontDesc; }
			set { this.ModifyFont(value); }
		}

		/// <summary>
		/// The text foreground color.
		/// </summary>
		public Gdk.Color ForeColor {
			get { return this.Style.Text(StateType.Normal); }
			set { this.ModifyText(StateType.Normal, value); }
		}

		/// <summary>
		/// The text background color.
		/// </summary>
		public Gdk.Color BackColor {
			get { return this.Style.Base(StateType.Normal); }
			set { this.ModifyBase(StateType.Normal, value); }
		}
		#endregion

		#region Constructors
		/// <summary>
		/// Constructor.
		/// </summary>
		public TerminalWindow() {
			WrapMode = WrapMode.WordChar;
			Editable = false;
		}
		#endregion

		#region Public Methods
		/// <summary>
		/// Get the word containing the non-whitespace character at 'idx'.
		/// </summary>
		public string GetWordAt(int idx) {
			var text = this.Buffer.Text;
			var textLength = text.Length;
			if (textLength == 0)
				return "";

			// clip to string edges
			if (idx >= textLength)
				idx = textLength - 1;
			if (idx < 0)
				idx = 0;

			// if the index is whitespace, return nothing
			if (Char.IsWhiteSpace(text[idx]))
				return "";

			// if we are on a non-whitespace character, search for the edges of the word containing it
			int start = idx;
			while (start > 0 && !Char.IsWhiteSpace(text[start - 1]))
				--start;
			int end = idx;
			while (end < (textLength - 1) && !Char.IsWhiteSpace(text[end + 1]))
				++end;
			++end;

			// return the token
			return text.Substring(start, end - start);
		}

		/// <summary>
		/// Clear the selection in the main window.
		/// </summary>
		public void ClearSelection() {
			TextIter begin, end;
			this.Buffer.GetSelectionBounds(out begin, out end);
			this.Buffer.SelectRange(begin, begin);
		}

		/// <summary>
		/// Get the main window selected text.
		/// </summary>
		public string GetSelectedText() {
			// if there is a selection, use it
			if (this.Buffer.HasSelection) {
				TextIter begin, end;
				this.Buffer.GetSelectionBounds(out begin, out end);
				return this.Buffer.GetText(begin, end, true).Trim();
			}

			// otherwise, get the token closest to the cursor
			if (!ReferenceEquals(PopupNearestWord, null)) {
				return PopupNearestWord;
			}

			// otherwise return the nearest word to the caret
			return GetWordAt(this.Buffer.CursorPosition);
		}

		/// <summary>
		/// Scroll to the end of the window, to follow most recent incoming data.
		/// </summary>
		public void ScrollToEnd() {
			this.Buffer.PlaceCursor(this.Buffer.EndIter);
			this.ScrollToIter(this.Buffer.EndIter, 0.0, false, 0.0, 0.0);
		}
		#endregion

		#region Public Events
		/// <summary>
		/// Raised when a key is pressed on the window.
		/// </summary>
		public event EventHandler<UnfilteredKeyPressEventArgs> UnfilteredKeyPress;
		#endregion

		#region Event Handlers
		/// <summary>
		/// Raised when a key is pressed on the window.
		/// </summary>
		protected override bool OnKeyPressEvent(Gdk.EventKey evnt) {
			PopupNearestWord = null;

			var e = new UnfilteredKeyPressEventArgs { Event = evnt };
			if (UnfilteredKeyPress != null) {
				UnfilteredKeyPress(this, e);
			}

			if (e.CancelEvent) {
				return true;
			}

			return base.OnKeyPressEvent(evnt);
		}

		/// <summary>
		/// Raised by a mouse button press event.
		/// </summary>
		protected override bool OnButtonPressEvent(Gdk.EventButton e) {
			if (e.Button == 3) {
				int x, y;
				this.GetPointer(out x, out y);
				var iter = GetIterAtLocation(x, y);
				Console.WriteLine("TerminalWindow.OnButtonPressEvent(...) right-clicked on pos: {0}, {1}", x, y);
				PopupNearestWord = GetWordAt(iter.Offset);
				if (!String.IsNullOrEmpty(PopupNearestWord)) {
					Console.WriteLine("TerminalWindow.OnButtonPressEvent(...) right-clicked on word: {0}", PopupNearestWord);
				}
			} else {
				PopupNearestWord = null;
			}

			// chain to base
			return base.OnButtonPressEvent(e);
		}
		#endregion
	}
}
