﻿/*
 *
 *
 *    FlXml.cs
 *
 *    Minimal XMLRPC interfaces to FlLog, FlRig.
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2014-2017 by Matthew K. Roberts, KK5JY.
 *    All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *    
 *
 */

using System;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Net;
using System.IO;
using System.Xml;
using KK5JY.Log;

namespace KK5JY.XMLRPC {
	/// <summary>
	/// Basic XMLRPC interface for fldigi programs.
	/// </summary>
	public abstract class FlXml {
		#region public Properties
		/// <summary>
		/// The URI of the remote endpoint.
		/// </summary>
		public string EndpointUri { get; protected set; }
		#endregion

		#region Utilities
		/// <summary>
		/// Format an XML-RPC request.
		/// </summary>
		private byte[] FormatRequest(string method, params string[] args) {
			using (var io = new MemoryStream()) {
				using (var w = new StreamWriter(io)) {
					w.WriteLine("<?xml version=\"1.0\" encoding=\"us-ascii\"?>");
					w.WriteLine("<methodCall><methodName>{0}</methodName><params>", method);
					foreach (var arg in args) {
						var escaped = arg ?? "";
						if (!String.IsNullOrEmpty(escaped))
							escaped = arg.Replace("&", "&amp;").Replace("<", "&lt;").Replace(">", "&gt;");
						w.WriteLine("<param><value>{0}</value></param>", escaped);
					}
					w.WriteLine("</params></methodCall>");
					w.Flush();
				}
				return io.ToArray();
			}
		}

		/// <summary>
		/// Execute the XML-RPC request.
		/// </summary>
		protected string Request(string method, params string[] args) {
			try {
				using (var wc = new WebClient()) {
					// render the request into XML(RPC)
					var request = FormatRequest(method, args);

					#if VERBOSE_DEBUG
					Debug.WriteLine(String.Format("DEBUG: FlXml.Request({0}, {1} args) BEGIN ==========", method, (args == null) ? 0 : args.Length));
					Debug.WriteLine(System.Text.Encoding.ASCII.GetString(request).Trim());
					Debug.WriteLine("DEBUG: FlXml.Request(...) ======= END ==========");
					#endif

					// send the request
					var response = wc.UploadData(EndpointUri, "POST", request);

					// parse the response XML
					var doc = new XmlDocument();
					using (var io = new MemoryStream(response)) {
						doc.Load(io);
						foreach (var element in doc.GetElementsByTagName("value").OfType<XmlElement>()) {
							return element.InnerText.Trim();
						}
					}
				}
			} catch (Exception ex) {
				// nop, but print to DEBUG for use in debugging
				Debug.WriteLine(
					String.Format("DEBUG: FlXml.Request({0}, ...) error: {1}: {2}",
						method,
						ex.GetType().Name,
						ex.Message));
			}

			// if the request failed, or didn't specify a return value, return null
			// NOTE: some commands return empty string (""), which is normal, and different from the case below
			return null;
		}
		#endregion
	}

	/// <summary>
	/// Basic FlRig interface.
	/// </summary>
	public class FlRig : FlXml {
		#region Public Methods
		/// <summary>
		/// Fetch the mode.
		/// </summary>
		public string GetMode() {
			return Request("rig.get_mode");
		}

		/// <summary>
		/// Fetch the frequency.
		/// </summary>
		public decimal GetFrequency() {
			var s = Request("rig.get_vfo");
			decimal d;
			if (Decimal.TryParse(s, out d))
				return d;
			return 0m;
		}
		#endregion

		#region Private Fields
		private string m_Host;
		private int m_Port;
		#endregion

		#region Constructors
		/// <summary>
		/// Create a new FlRig remote object.
		/// </summary>
		/// <param name="host">The host name.</param>
		/// <param name="port">The port number.</param>
		public FlRig(string host = "localhost", int port = 12345) {
			m_Host = host;
			m_Port = port;
			EndpointUri = String.Format("http://{0}:{1}/RPC2", m_Host, m_Port);
		}
		#endregion
	}

	/// <summary>
	/// Basic FlLog interface.
	/// </summary>
	public class FlLog : FlXml {
		#region Public Methods
		/// <summary>
		/// Fetch the frequency.
		/// </summary>
		public bool DupeCheck(string call, string mode, string freq_kHz) {
			var s = Request("log.check_dup", call, mode, "", freq_kHz, "", "");
			bool result = false;
			if (Boolean.TryParse(s, out result)) {
				return result;
			}
			return false;
		}

		/// <summary>
		/// Get an entry
		/// </summary>
		public AdifRecord GetADIF(string call) {
			var s = Request("log.get_record", call);
			if (String.IsNullOrEmpty(s) || String.Equals(s, "NO_RECORD")) {
				return null;
			}
			return AdifRecord.Parse(s);
		}

		/// <summary>
		/// Add an entry
		/// </summary>
		public string AddADIF(AdifRecord adif) {
			return Request("log.add_record", adif.ToString());
		}
		#endregion

		#region Private Fields
		private string m_Host;
		private int m_Port;
		#endregion

		#region Constructors
		/// <summary>
		/// Create a new FlLog remote object.
		/// </summary>
		/// <param name="host">The host name.</param>
		/// <param name="port">The port number.</param>
		public FlLog(string host = "localhost", int port = 8421) {
			m_Host = host;
			m_Port = port;
			EndpointUri = String.Format("http://{0}:{1}/RPC2", m_Host, m_Port);
		}
		#endregion
	}
}
