﻿/*
 *
 *
 *    MacroExpansion.cs
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2017 by Matthew K. Roberts, KK5JY. All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *    
 *
 */

using System;
using System.Collections.Generic;

namespace KK5JY.SMI {
	/// <summary>
	/// Possible macro value request types.
	/// </summary>
	public enum MacroKeys {
		Call,
		MyCall,
		RxExch,
		TxExch,
		LogCount,
		LogNext,
	}

	public sealed class MacroExpansionField : EventArgs {
		/// <summary>
		/// The requested key.
		/// </summary>
		public MacroKeys Key { get; private set; }

		/// <summary>
		/// The value provided by the handler.
		/// </summary>
		public string Value { get; set; }

		public MacroExpansionField(MacroKeys key) {
			this.Key = key;
		}
	}

	/// <summary>
	/// Expansion request delegate.
	/// </summary>
	public delegate void MacroExpansionFieldRequest(MacroExpansionField f);

	/// <summary>
	/// Macro expansion table.
	/// </summary>
	public class MacroExpansion {
		/// <summary>
		/// Raised to request a macro value.
		/// </summary>
		public event EventHandler<MacroExpansionField> FieldRequest;

		/// <summary>
		/// Create a new macro expander.
		/// </summary>
		public MacroExpansion() {
			// nop
		}

		/// <summary>
		/// Expand macros in a string.  Repeat the expansion as long as there are symbols to expand.
		/// </summary>
		/// <param name="s">The text containing macro variables.</param>
		/// <returns>The expanded string</returns>
		public string ExpandRecursive(string s) {
			bool modified = false;
			do {
				var newText = Expand(s);
				modified = !newText.Equals(s, StringComparison.InvariantCultureIgnoreCase);
				s = newText;
			} while (modified);
			return s;
		}

		/// <summary>
		/// Expand macros in a string.
		/// </summary>
		/// <param name="s">The text containing macro variables.</param>
		/// <returns>The expanded string</returns>
		public string Expand(string s) {
			var macros = GetVariables(s);
			if (macros.Count == 0) return s;
			if (FieldRequest == null) return s;

			foreach (var token in macros) {
				// remove the format specifier, if any
				var realToken = token;
				string formatSpec = null;
				if (token.Contains(":")) {
					var parts = realToken.Split(new char[] { ':' }, 2, StringSplitOptions.RemoveEmptyEntries);
					if (parts.Length == 2) {
						realToken = parts[0].Trim();
						formatSpec = parts[1].Trim();
					}
				}

				// call out for the replacement text
				var type = MacroKeys.Call;
				switch (realToken.ToUpper().Trim()) {
					case "CALL": type = MacroKeys.Call; break;
					case "MYCALL": type = MacroKeys.MyCall; break;
					case "RXEXCH": type = MacroKeys.RxExch; break;
					case "TXEXCH": type = MacroKeys.TxExch; break;
					case "LOGCOUNT": type = MacroKeys.LogCount; break;
					case "LOGNEXT": type = MacroKeys.LogNext; break;
					default: continue;
				}
				var req = new MacroExpansionField(type);
				FieldRequest(this, req);

				// if a format specifier was provided
				if (!String.IsNullOrEmpty(formatSpec)) {
					switch (type) {
						// and if the macro type is one of the log counters
						case MacroKeys.LogCount:
						case MacroKeys.LogNext: {
							Int32 intValue;
							// and if the provided value is a valid integer
							if (Int32.TryParse(req.Value, out intValue)) {
								// replace the value with a reformatted version
								var formatString = "{0:" + formatSpec + "}";
								try {
									// (try to) format the number using the string specified
									req.Value = String.Format(formatString, intValue);

									// if the format specifier was bogus but functional, just put back the integer
									if (req.Value == formatSpec) {
										req.Value = intValue.ToString();
									}
								} catch {
									// if the format string was invalid, just put back the original value and move on
									req.Value = intValue.ToString();
								}
							}
						} break;
					}
				}

				//
				//  TODO: This replacement assumes that the '${' sequence has no intervening whitespace.
				//        Replace this with code that does not have this limitation.
				//

				// do the replacement
				s = s.Replace("${" + token + "}", req.Value ?? "");
			}

			return s;
		}

		/// <summary>
		/// Scan the string to determine the variable names used.
		/// </summary>
		private LinkedList<string> GetVariables(string s) {
			int idx = 0;
			int state = 0;
			var tokens = new LinkedList<string>();
			string token = "";

			for (idx = 0; idx != s.Length; ++idx) {
				switch (s[idx]) {
					//
					//  state transitions
					//
					case '$': {
						if (state == 0)
							state = 1;
					} continue;
					case '{': {
						if (state == 1)
							state = 2;
					} continue;
					case '}': {
						if (state == 2)
							state = 0;
						if (!String.IsNullOrEmpty(token)) {
							tokens.AddLast(token);
							token = "";
						}
					} continue;

					//
					//  this is the main handler
					//
					default: {
						if (state == 2) {
							token += s[idx];
						}
					} break;
				}
			}

			return tokens;
		}
	}
}
