﻿/*
 *
 *
 *    UnitTests.cs
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2017 by Matthew K. Roberts, KK5JY. All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *    
 *
 */

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;

using Microsoft.VisualStudio.TestTools.UnitTesting;

using KK5JY.SMI;
using KK5JY.Log;

namespace UnitTests {
	[TestClass]
	public class UnitTests {
		#region Macro Expansion Unit Tests
		/// <summary>
		/// This is the test handler -- it returns values in response to request for variables, and emulates
		/// the behavior of the main application.
		/// </summary>
		private void FieldRequestHandler(object sender, MacroExpansionField e) {
			switch (e.Key) {
				case MacroKeys.Call: e.Value = "N5OSL"; break;
				case MacroKeys.MyCall: e.Value = "KK5JY"; break;
				case MacroKeys.LogNext: e.Value = "6"; break;
				case MacroKeys.LogCount: e.Value = "5"; break;
			}
		}

		/// <summary>
		/// Test single variable expansion with some spaces.
		/// </summary>
		[TestMethod]
		public void TestMacroSuccess1() {
			var testString = "CQ CQ CQ de ${MYCALL} ${  MYCALL } ${ MYCALL}";
			var expected = "CQ CQ CQ de KK5JY KK5JY KK5JY";
			var m = new MacroExpansion();
			m.FieldRequest += new EventHandler<MacroExpansionField>(FieldRequestHandler);
			var result = m.Expand(testString);
			Assert.AreEqual(expected, result);
		}

		/// <summary>
		/// Test more variables with more spaces.
		/// </summary>
		[TestMethod]
		public void TestMacroSuccess2() {
			var testString = "${CALL} ${CALL } ${ CALL } de ${ MYCALL } ${  MYCALL } ${ MYCALL}";
			var expected = "N5OSL N5OSL N5OSL de KK5JY KK5JY KK5JY";
			var m = new MacroExpansion();
			m.FieldRequest += new EventHandler<MacroExpansionField>(FieldRequestHandler);
			var result = m.Expand(testString);
			Assert.AreEqual(expected, result);
		}

		/// <summary>
		/// Test numeric expansion.
		/// </summary>
		[TestMethod]
		public void TestMacroSuccess3() {
			var testString = "${CALL} 599-${LOGNEXT} ${ MYCALL}";
			var expected = "N5OSL 599-6 KK5JY";
			var m = new MacroExpansion();
			m.FieldRequest += new EventHandler<MacroExpansionField>(FieldRequestHandler);
			var result = m.Expand(testString);
			Assert.AreEqual(expected, result);
		}

		/// <summary>
		/// Test numeric expansion with formatting.
		/// </summary>
		[TestMethod]
		public void TestMacroSuccess4() {
			var testString = "${CALL} 599-${LOGNEXT:00} ${ MYCALL}";
			var expected = "N5OSL 599-06 KK5JY";
			var m = new MacroExpansion();
			m.FieldRequest += new EventHandler<MacroExpansionField>(FieldRequestHandler);
			var result = m.Expand(testString);
			Assert.AreEqual(expected, result);
		}

		/// <summary>
		/// Test numeric expansion with formatting.
		/// </summary>
		[TestMethod]
		public void TestMacroSuccess5() {
			var testString = "${CALL} 599-${LOGNEXT:000} ${ MYCALL}";
			var expected = "N5OSL 599-006 KK5JY";
			var m = new MacroExpansion();
			m.FieldRequest += new EventHandler<MacroExpansionField>(FieldRequestHandler);
			var result = m.Expand(testString);
			Assert.AreEqual(expected, result);
		}

		/// <summary>
		/// Test numeric expansion with invalid formatting.
		/// </summary>
		[TestMethod]
		public void TestMacroSuccess6() {
			var testString = "${CALL} 599-${LOGNEXT:BLAH} ${ MYCALL}";
			var expected = "N5OSL 599-6 KK5JY";
			var m = new MacroExpansion();
			m.FieldRequest += new EventHandler<MacroExpansionField>(FieldRequestHandler);
			var result = m.Expand(testString);
			Assert.AreEqual(expected, result);
		}

		/// <summary>
		/// Test numeric expansion with alternative formatting.
		/// </summary>
		[TestMethod]
		public void TestMacroSuccess7() {
			var testString = "${CALL} 599-${LOGNEXT:D6} ${ MYCALL}";
			var expected = "N5OSL 599-000006 KK5JY";
			var m = new MacroExpansion();
			m.FieldRequest += new EventHandler<MacroExpansionField>(FieldRequestHandler);
			var result = m.Expand(testString);
			Assert.AreEqual(expected, result);
		}

		/// <summary>
		/// Test numeric expansion with alternative formatting and redundant spaces.
		/// </summary>
		[TestMethod]
		public void TestMacroSuccess8() {
			var testString = "${CALL} 599-${ LOGNEXT : D6 } ${MYCALL}";
			var expected = "N5OSL 599-000006 KK5JY";
			var m = new MacroExpansion();
			m.FieldRequest += new EventHandler<MacroExpansionField>(FieldRequestHandler);
			var result = m.Expand(testString);
			Assert.AreEqual(expected, result);
		}

		/// <summary>
		/// Test expansion when there are no variables.
		/// </summary>
		[TestMethod]
		public void TestMacroNothingToDo() {
			var testString = "CQ CQ CQ de ME ME ME";
			var expected = testString;
			var m = new MacroExpansion();
			m.FieldRequest += new EventHandler<MacroExpansionField>(FieldRequestHandler);
			var result = m.Expand(testString);
			Assert.AreEqual(expected, result);
		}

		/// <summary>
		/// Test all kinds of invalid variables.
		/// </summary>
		[TestMethod]
		public void TestMacroBadMacros() {
			var testString = " CQ ${CQ} CQ de ${ ME } ME ME ${} ${ } $$ ${ ${ }";
			var expected = testString;
			var m = new MacroExpansion();
			m.FieldRequest += new EventHandler<MacroExpansionField>(FieldRequestHandler);
			var result = m.Expand(testString);
			Assert.AreEqual(expected, result);
		}
		#endregion

		#region ADIF File I/O Tests
		private readonly string AdifPath = Path.Combine(Environment.CurrentDirectory, "UnitTests.adif");
		private AdifRecord TestRecord;
		private void WriteAdif() {
			// construct the record
			TestRecord = new AdifRecord {
				Call = "KK5JY",
				Mode = "CW",
				Freq = "14.050",
				RxString = "1E OK",
				TxString = "5A MDC",
			};
			TestRecord.SetQsoTime(DateTime.Now);

			// write the file with one record
			AdifFile file = new AdifFile(AdifPath);
			file.AddRecord(TestRecord);
		}

		[TestMethod]
		public void WriteAdifClean() {
			// write the ADIF
			WriteAdif();

			// dump file and clean up
			if (File.Exists(AdifPath)) {
				using (var f = File.OpenRead(AdifPath)) {
					using (var sw = new StreamReader(f)) {
						var data = sw.ReadToEnd();
						Debug.WriteLine("DEBUG ====== ADIF DATA ======");
						Debug.Write(data);
						Debug.WriteLine("DEBUG ====== ADIF DATA ======");
					}
				}
				File.Delete(AdifPath);
			}
		}

		[TestMethod]
		public void ReadAdifClean() {
			WriteAdif();

			var file = new AdifFile(AdifPath);
			int count = 0;
			foreach (var item in file.GetAllRecords()) {
				++count;
				Assert.AreEqual(TestRecord.Call, item.Call);
				Assert.AreEqual(TestRecord.Freq, item.Freq);
				Assert.AreEqual(TestRecord.Mode, item.Mode);
				Assert.AreEqual(TestRecord.DateOn, item.DateOn);
				Assert.AreEqual(TestRecord.TimeOn, item.TimeOn);
				Assert.AreEqual(TestRecord.TimeOff, item.TimeOff);
				Assert.AreEqual(TestRecord.RxString, item.RxString);
				Assert.AreEqual(TestRecord.TxString, item.TxString);
			}
			Assert.AreEqual(1, count);

			count = 0;
			foreach (var item in file.GetRecords("N0BDY", 20)) {
				++count;
			}
			Assert.AreEqual(0, count);

			count = 0;
			foreach (var item in file.GetRecords(TestRecord.Call, 30)) {
				++count;
			}
			Assert.AreEqual(0, count);

			count = 0;
			foreach (var item in file.GetRecords(TestRecord.Call, 20)) {
				++count;
			}
			Assert.AreEqual(1, count);

			count = 0;
			foreach (var item in file.GetRecords(TestRecord.Call, 20, "CW")) {
				++count;
			}
			Assert.AreEqual(1, count);

			count = 0;
			foreach (var item in file.GetRecords(TestRecord.Call, 20, "PSK31")) {
				++count;
			}
			Assert.AreEqual(0, count);

			// dump file and clean up
			if (File.Exists(AdifPath)) {
				File.Delete(AdifPath);
			}
		}

		[TestMethod]
		public void DeleteAdifRecord() {
			WriteAdif();

			var file = new AdifFile(AdifPath);
			int count = 0;
			foreach (var item in file.GetAllRecords()) {
				++count;
				Assert.AreEqual(TestRecord.Call, item.Call);
				Assert.AreEqual(TestRecord.Freq, item.Freq);
				Assert.AreEqual(TestRecord.Mode, item.Mode);
				Assert.AreEqual(TestRecord.DateOn, item.DateOn);
				Assert.AreEqual(TestRecord.TimeOn, item.TimeOn);
				Assert.AreEqual(TestRecord.TimeOff, item.TimeOff);
				Assert.AreEqual(TestRecord.RxString, item.RxString);
				Assert.AreEqual(TestRecord.TxString, item.TxString);
			}
			Assert.AreEqual(1, count);

			count = 0;
			foreach (var item in file.GetRecords(TestRecord.Call, 20, "CW")) {
				++count;
			}
			Assert.AreEqual(1, count);

			// delete one record
			file.DeleteRecord(TestRecord);
			count = 0;
			foreach (var item in file.GetRecords(TestRecord.Call, 20, "CW")) {
				++count;
			}
			Assert.AreEqual(0, count);

			// dump file and clean up
			if (File.Exists(AdifPath)) {
				File.Delete(AdifPath);
			}
		}

		[TestMethod]
		public void DeleteAdifRecordVariousMiddle() {
			// construct the records
			TestRecord = new AdifRecord {
				Call = "KK5JY",
				Mode = "CW",
				Freq = "14.050",
				RxString = "1E OK",
				TxString = "5A MDC",
			};
			TestRecord.SetQsoTime(DateTime.Now);
			var record1 = new AdifRecord {
				Call = "KK5JY",
				Mode = "CW",
				Freq = "14.050",
				RxString = "1E OK",
				TxString = "5A MDC",
			};
			record1.SetQsoTime(DateTime.Now + TimeSpan.FromSeconds(100));
			var record2 = new AdifRecord {
				Call = "N5OSL",
				Mode = "PSK31",
				Freq = "14.070",
				RxString = "2A MT",
				TxString = "5D EPA",
			};
			record2.SetQsoTime(DateTime.Now + TimeSpan.FromSeconds(200));
			var record3 = new AdifRecord {
				Call = "AB5CD",
				Mode = "PSK63",
				Freq = "7.035",
				RxString = "1E OK",
				TxString = "5A MDC",
			};
			record3.SetQsoTime(DateTime.Now + TimeSpan.FromSeconds(300));
			var record4 = new AdifRecord {
				Call = "KK5JY",
				Mode = "CW",
				Freq = "14.050",
				RxString = "1E OK",
				TxString = "5A MDC",
			};
			record4.SetQsoTime(DateTime.Now + TimeSpan.FromSeconds(500));

			var file = new AdifFile(AdifPath);
			file.AddRecord(record1);
			file.AddRecord(record2);
			file.AddRecord(TestRecord);
			file.AddRecord(record3);
			file.AddRecord(record4);

			// make sure all records were inserted
			int count = file.Count;
			Assert.AreEqual(5, count);

			// delete one record
			file.DeleteRecord(TestRecord);
			count = file.Count;
			Assert.AreEqual(4, count);

			// make sure it is gone
			count = 0;
			foreach (var item in file.GetRecords(TestRecord.Call, 20, TestRecord.Mode)) {
				if (item.Equals(TestRecord))
					++count;
			}
			Assert.AreEqual(0, count);

			// dump file and clean up
			if (File.Exists(AdifPath)) {
				File.Delete(AdifPath);
			}
		}

		[TestMethod]
		public void DeleteAdifRecordVariousStart() {
			// construct the records
			TestRecord = new AdifRecord {
				Call = "KK5JY",
				Mode = "CW",
				Freq = "14.050",
				RxString = "1E OK",
				TxString = "5A MDC",
			};
			TestRecord.SetQsoTime(DateTime.Now);
			var record1 = new AdifRecord {
				Call = "KK5JY",
				Mode = "CW",
				Freq = "14.050",
				RxString = "1E OK",
				TxString = "5A MDC",
			};
			record1.SetQsoTime(DateTime.Now + TimeSpan.FromSeconds(100));
			var record2 = new AdifRecord {
				Call = "N5OSL",
				Mode = "PSK31",
				Freq = "14.070",
				RxString = "2A MT",
				TxString = "5D EPA",
			};
			record2.SetQsoTime(DateTime.Now + TimeSpan.FromSeconds(200));
			var record3 = new AdifRecord {
				Call = "AB5CD",
				Mode = "PSK63",
				Freq = "7.035",
				RxString = "1E OK",
				TxString = "5A MDC",
			};
			record3.SetQsoTime(DateTime.Now + TimeSpan.FromSeconds(300));
			var record4 = new AdifRecord {
				Call = "KK5JY",
				Mode = "CW",
				Freq = "14.050",
				RxString = "1E OK",
				TxString = "5A MDC",
			};
			record4.SetQsoTime(DateTime.Now + TimeSpan.FromSeconds(500));

			var file = new AdifFile(AdifPath);
			file.AddRecord(TestRecord);
			file.AddRecord(record1);
			file.AddRecord(record2);
			file.AddRecord(record3);
			file.AddRecord(record4);

			// make sure all records were inserted
			int count = file.Count;
			Assert.AreEqual(5, count);

			// delete one record
			file.DeleteRecord(TestRecord);
			count = file.Count;
			Assert.AreEqual(4, count);

			// make sure it is gone
			count = 0;
			foreach (var item in file.GetRecords(TestRecord.Call, 20, TestRecord.Mode)) {
				if (item.Equals(TestRecord))
					++count;
			}
			Assert.AreEqual(0, count);

			// dump file and clean up
			if (File.Exists(AdifPath)) {
				File.Delete(AdifPath);
			}
		}

		[TestMethod]
		public void DeleteAdifRecordVariousEnd() {
			// construct the records
			TestRecord = new AdifRecord {
				Call = "KK5JY",
				Mode = "CW",
				Freq = "14.050",
				RxString = "1E OK",
				TxString = "5A MDC",
			};
			TestRecord.SetQsoTime(DateTime.Now);
			var record1 = new AdifRecord {
				Call = "KK5JY",
				Mode = "CW",
				Freq = "14.050",
				RxString = "1E OK",
				TxString = "5A MDC",
			};
			record1.SetQsoTime(DateTime.Now + TimeSpan.FromSeconds(100));
			var record2 = new AdifRecord {
				Call = "N5OSL",
				Mode = "PSK31",
				Freq = "14.070",
				RxString = "2A MT",
				TxString = "5D EPA",
			};
			record2.SetQsoTime(DateTime.Now + TimeSpan.FromSeconds(200));
			var record3 = new AdifRecord {
				Call = "AB5CD",
				Mode = "PSK63",
				Freq = "7.035",
				RxString = "1E OK",
				TxString = "5A MDC",
			};
			record3.SetQsoTime(DateTime.Now + TimeSpan.FromSeconds(300));
			var record4 = new AdifRecord {
				Call = "KK5JY",
				Mode = "CW",
				Freq = "14.050",
				RxString = "1E OK",
				TxString = "5A MDC",
			};
			record4.SetQsoTime(DateTime.Now + TimeSpan.FromSeconds(500));

			var file = new AdifFile(AdifPath);
			file.AddRecord(record1);
			file.AddRecord(record2);
			file.AddRecord(record3);
			file.AddRecord(record4);
			file.AddRecord(TestRecord);

			// make sure all records were inserted
			int count = file.Count;
			Assert.AreEqual(5, count);

			// delete one record
			file.DeleteRecord(TestRecord);
			count = file.Count;
			Assert.AreEqual(4, count);

			// make sure it is gone
			count = 0;
			foreach (var item in file.GetRecords(TestRecord.Call, 20, TestRecord.Mode)) {
				if (item.Equals(TestRecord))
					++count;
			}
			Assert.AreEqual(0, count);

			// dump file and clean up
			if (File.Exists(AdifPath)) {
				File.Delete(AdifPath);
			}
		}
		#endregion
	}
}
